/* * (C) Copyright IBM Corporation 1991, 2015. */
/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA


 Author: IBM Corporation
 This module is part of the IBM (R) Rational (R) ClearCase (R)
 Multi-version file system (MVFS).
 For support, please visit http://www.ibm.com/software/support

*/
/* mvfs_mioctl.c */

#include "mvfs_systm.h"
#include <tbs_base.h>
#include <tbs_errno.h>
#include <credutl_kernel.h>
#include "mvfs.h"
#include "mvfs_dnc.h"
#include "mvfs_copy.h"

/* MVFS ioctl subroutines */
/* All subroutines must have the following prototype. */
typedef int (*mvfs_ioctl_subr_t)(
    mvfs_viewroot_type_t vrtype,
    mvfscmd_block_t *data,
    CALL_DATA_T *cd,
    MVFS_CALLER_INFO *callinfo
);

#define MVFS_IOCTL_HANDLER(ioctl_suffix) \
    STATIC int \
    mvfs_ioctl_##ioctl_suffix(\
        mvfs_viewroot_type_t vrtype, \
        mvfscmd_block_t *data, \
        CALL_DATA_T *cd, \
        MVFS_CALLER_INFO *callinfo \
    )

MVFS_IOCTL_HANDLER(get_viewinfo);
MVFS_IOCTL_HANDLER(xstat);
MVFS_IOCTL_HANDLER(get_clrname);
MVFS_IOCTL_HANDLER(get_vfh);
MVFS_IOCTL_HANDLER(ioinval);
MVFS_IOCTL_HANDLER(audit_ioctl);
MVFS_IOCTL_HANDLER(mkviewtag);
MVFS_IOCTL_HANDLER(rmviewtag);
MVFS_IOCTL_HANDLER(get_viewaddr);
MVFS_IOCTL_HANDLER(get_vobinfo);
MVFS_IOCTL_HANDLER(get_viewtag_dir);
MVFS_IOCTL_HANDLER(change_mtype);
MVFS_IOCTL_HANDLER(get_vxsuffix);
MVFS_IOCTL_HANDLER(get_cache_enb);
MVFS_IOCTL_HANDLER(set_cache_enb);
MVFS_IOCTL_HANDLER(flush_cache);
MVFS_IOCTL_HANDLER(read_dnc);
MVFS_IOCTL_HANDLER(get_rcsid);
MVFS_IOCTL_HANDLER(get_sccsid);
MVFS_IOCTL_HANDLER(get_loginfo);
MVFS_IOCTL_HANDLER(set_loginfo);
MVFS_IOCTL_HANDLER(get_stats);
MVFS_IOCTL_HANDLER(set_procview);
MVFS_IOCTL_HANDLER(get_procviewinfo);
MVFS_IOCTL_HANDLER(get_xattr);
MVFS_IOCTL_HANDLER(set_xattr);
MVFS_IOCTL_HANDLER(mount);
MVFS_IOCTL_HANDLER(unmount);
MVFS_IOCTL_HANDLER(rmallviewtags);
MVFS_IOCTL_HANDLER(unmountall);
MVFS_IOCTL_HANDLER(get_poolmaps);

/* This is a generic routine used for ioctls that are no longer in use or are
 * not supported on a particular platform.
 */
MVFS_IOCTL_HANDLER(unsupported);

MVFS_IOCTL_HANDLER(abort);
MVFS_IOCTL_HANDLER(exportviewtag);
MVFS_IOCTL_HANDLER(unexportviewtag);
MVFS_IOCTL_HANDLER(zero_stats);
MVFS_IOCTL_HANDLER(get_cache_usage);
MVFS_IOCTL_HANDLER(set_cache_sizes);
MVFS_IOCTL_HANDLER(get_view_stats);
MVFS_IOCTL_HANDLER(zero_view_stats);
MVFS_IOCTL_HANDLER(sidhost_credmapping);
MVFS_IOCTL_HANDLER(delete_sidhost_credmapping);
MVFS_IOCTL_HANDLER(get_viewtag_export);
MVFS_IOCTL_HANDLER(get_gfsinfo);
MVFS_IOCTL_HANDLER(set_vobrt_vfsmnt);
MVFS_IOCTL_HANDLER(get_cache_sizes);

MVFS_IOCTL_HANDLER(compute_cache_defaults);

MVFS_IOCTL_HANDLER(enable_pview_stats);
MVFS_IOCTL_HANDLER(disable_pview_stats);

/* Utility and helper functions */

/* mvfs ioctl lookup flags
 *
 * These are internal flags used by mvfs_ioctl_lookup. The user visible lookup
 * flags in mfs_ioctl.h (MVFS_CMD_PN_xxx) are translated to these.
 */
typedef enum {
    MVFS_IL_FLAG_NONE = 0x0, /* No flags */
    MVFS_IL_FLAG_NO_FOLLOW_SYMLINK = 0x1,
    MVFS_IL_FLAG_CS_LOOKUP = 0x2,
    MVFS_IL_FLAG_PN_IS_VIEWTAG = 0x4,
    MVFS_IL_FLAG_NB_LOOKUP = 0x8
} mvfs_il_flags_t;

/* Translates the MVFS_CMD_PN_XXX flags to the internal versions used by
 * mvfs_ioctl_lookup.
 */
STATIC mvfs_il_flags_t
mvfs_ioctl_lookup_pnflags_to_ilflags(
    mvfs_cmd_pn_flags_t pnflags
);

#define mvfs_get_ilflags_from_cb(cmdblk) \
    mvfs_ioctl_lookup_pnflags_to_ilflags(MCB_PNFLAGS((cmdblk)))

STATIC int MVFS_NOINLINE
mvfs_ioctl_lookup(
    mvfs_viewroot_type_t vrtype,
    mfs_strbufpn_t pnp,
    mvfs_il_flags_t flags,
    VNODE_T **vpp,
    VNODE_T **xvpp,
    CALL_DATA_T *cd,
    MVFS_CALLER_INFO *callinfo
);

STATIC int MVFS_NOINLINE
mvfs_copyout_vobinfo(
    mvfs_vobinfo_t *vip,
    struct mfs_mntinfo *mmi,
    tbs_boolean_t unique ,
    mvfscmd_block_t *cmdblk,
    MVFS_CALLER_INFO *callinfo
);

STATIC int MVFS_NOINLINE
mvfs_copyout_viewinfo(
    mvfs_viewinfo_t *vip,
    VNODE_T *vp
);

STATIC int MVFS_NOINLINE
mvfs_do_inval(
    mvfs_viewroot_type_t vrtype,
    tbs_oid_t *replica_oid_p,
    mvfs_ioinval_t *invp,
    VNODE_T *cdir,
    VNODE_T *vp,
    VFS_T *vfsp,
    tbs_boolean_t inval_local_only,
    CALL_DATA_T *cd
);

STATIC int MVFS_NOINLINE
mvfs_mkviewtag_subr(
    mvfs_viewroot_type_t vrtype,
    mvfscmd_block_t *data,
    mvfs_mkviewtag_info_ex_t *vtp_ex,
    CALL_DATA_T *cd,
    MVFS_CALLER_INFO *callinfo
);

STATIC void
mvfs_addup_clntstat(
    mvfs_stats_data_t *sdp,
    mvfs_stats_data_t *total_sdp
);

STATIC void
mvfs_addup_mnstat(
    mvfs_stats_data_t *sdp,
    mvfs_stats_data_t *total_sdp
);

STATIC void
mvfs_addup_clearstat(
    mvfs_stats_data_t *sdp,
    mvfs_stats_data_t *total_sdp
);

STATIC void
mvfs_addup_rvcstat(
    mvfs_stats_data_t *sdp,
    mvfs_stats_data_t *total_sdp
);

STATIC void
mvfs_addup_dncstat(
    mvfs_stats_data_t *sdp,
    mvfs_stats_data_t *total_sdp
);

STATIC void
mvfs_addup_acstat(
    mvfs_stats_data_t *sdp,
    mvfs_stats_data_t *total_sdp
);

STATIC void
mvfs_addup_rlstat(
    mvfs_stats_data_t *sdp,
    mvfs_stats_data_t *total_sdp
);

STATIC void
mvfs_addup_austat(
    mvfs_stats_data_t *sdp,
    mvfs_stats_data_t *total_sdp
);

STATIC void
mvfs_addup_eacstat(
    mvfs_stats_data_t *sdp,
    mvfs_stats_data_t *total_sdp
);

STATIC void
mvfs_addup_vnopcnt(
    mvfs_stats_data_t *sdp,
    mvfs_stats_data_t *total_sdp
);

STATIC void
mvfs_addup_vfsopcnt(
    mvfs_stats_data_t *sdp,
    mvfs_stats_data_t *total_sdp
);

STATIC void
mvfs_addup_viewopcnt(
    mvfs_stats_data_t *sdp,
    mvfs_stats_data_t *total_sdp
);

STATIC void
mvfs_addup_viewophist(
    mvfs_stats_data_t *sdp,
    mvfs_stats_data_t *total_sdp
);

STATIC void
mvfs_addup_viewoptime(
    mvfs_stats_data_t *sdp,
    mvfs_stats_data_t *total_sdp
);

STATIC void
mvfs_add_times(
      timestruc_t *sdp_time,
      timestruc_t *percpu_time
);

/*
 * MVFS_IOCTL_COPYIN - copyin the ioctl command block and lookup pathname
 * returning vnode ptr if command takes a pathname.
 *
 * NOTE: The ioctl handlers are responsible for copying in/out
 *       data specific to the ioctl command.
 */

int
mvfs_ioctl_copyin(
    mvfscmd_block_t *data,
    mvfscmd_block_t *iocbufp,
    caddr_t *infop,
    MVFS_CALLER_INFO *callinfo
)
{
    int  error = 0;

    /* Check for no buffer */

    if (iocbufp == NULL) {
        return ENOMEM;
    }

    return CopyInMvfscmd_block((caddr_t)data, iocbufp, callinfo);
}

STATIC mvfs_il_flags_t
mvfs_ioctl_lookup_pnflags_to_ilflags(
    mvfs_cmd_pn_flags_t pnflags
)
{
    mvfs_il_flags_t ilflags = MVFS_IL_FLAG_NONE;

    if (pnflags & MVFS_CMD_PN_NOFOLLOW_SYMLINK) {
        ilflags |= MVFS_IL_FLAG_NO_FOLLOW_SYMLINK;
    }

    if (pnflags & MVFS_CMD_PN_CS_LOOKUP) {
        ilflags |= MVFS_IL_FLAG_CS_LOOKUP;
    }

    return ilflags;
}

/* mvfs_ioctl_lookup
 * Lookups up a set of vnodes for use in an ioctl.
 *
 * Inputs:
 * pnp - A pointer to a strbufpn with the path to lookup.
 *  - NOTE: A copyin is performed on the strbufpn.
 * flags - A combination of MVFS_CMD_PN_* flags.
 * cd - Call data.
 * callinfo - Caller info.
 *
 * Outputs:
 * vpp - A vnode resulting from the path lookup.
 * xvpp - A vnode representing what's "underneath" vpp.
 *  - If vpp is a VOBCLAS node, xvpp may be a VOBRTCLAS node or cleartext.
 *  - If vpp is a LOOPCLAS node, xvpp is the covered node.
 *  - In all other cases, xvpp is set to NULL.
 *
 * NOTE: There is one exception to the above output rules when
 * MVFS_IL_FLAG_NB_LOOKUP is passed. In this case, if vpp is a VOBRTCLAS node,
 * xvpp is NULL.
 *
 * Results:
 * 0 if successful, standard unix errno otherwise.
 */
int
mvfs_ioctl_lookup(
    mvfs_viewroot_type_t vrtype,
    mfs_strbufpn_t pnp,
    mvfs_il_flags_t flags,
    VNODE_T **vpp,
    VNODE_T **xvpp,
    CALL_DATA_T *cd,
    MVFS_CALLER_INFO *callinfo
)
{
    int error = 0;
    mvfs_thread_t *mth;
    SYMFOLLOW_T lookflag;
    int extra_lookflag;
    VNODE_T *xvp = NULL;
    CALL_DATA_T *ncd = NULL;
    VNODE_T *vp = NULL;
    CLR_VNODE_T *cvp = NULL;
    CLR_VNODE_T *rcvp = NULL;   /* Avoid a compiler warning by initializing. */
    char *pn = NULL;

    MFS_CHKSP(STK_LOOKUPNAME);

    /* Clear return pointers in case of errors. */
    if (vpp == NULL && xvpp == NULL) {
        return EINVAL;
    }

    if (vpp != NULL) {
        *vpp = NULL;
    }

    if (xvpp != NULL) {
        *xvpp = NULL;
    }

    if (pnp.s == NULL) {
        return EINVAL;
    }

    /* Before we inhibit auditing and allocate memory for creds, check to see
     * if we're being asked to lookup a viewtag. If so, do the lookup and
     * return.
     */
    if (flags & MVFS_IL_FLAG_PN_IS_VIEWTAG) {
        if ((error = mfs_copyin_strbufpn(pnp, &pn)) != 0) {
            return error;
        }

        if ((error = mvfs_viewtaglookup(vrtype, pn, vpp, cd)) != 0) {
            mvfs_log(MFS_LOG_DEBUG,
                     "mvfs_ioctl_lookup: view tag lookup for '%s' failed, "
                     "error = %d\n",
                     (pn == NULL ? "NULL" : pn),
                     error);
        }

        STRFREE(pn);

        return error;
    } else {
        /* We're looking up an actual path.
         * NOTE: LOOKUP_FOR_IOCTL will take care of copying in the path for us.
         */
        pn = pnp.s;
    }

    ncd = MVFS_ALLOC_SUBSTITUTE_CRED(cd, NULL);

    if (!MVFS_SUBSTITUTE_CRED_IS_VALID(ncd)) {
        return ENOMEM;
    }

    if (flags & MVFS_IL_FLAG_NO_FOLLOW_SYMLINK) {
        lookflag = NO_FOLLOW;
    } else {
        lookflag = FOLLOW_LINK;
    }

    extra_lookflag = MVFS_NB_LOOKUP | MVFS_RO_LOOKUP;

    if (!(flags & MVFS_IL_FLAG_CS_LOOKUP)) {
        extra_lookflag |= MVFS_CI_LOOKUP;
    }

    /*
     * INHIBIT auditing of lookups in ioctl's.  They
     * can cause stack overflows on some systems, and
     * we don't want all this meta-junk in the audit.
     */

    mth = MVFS_MYTHREAD(cd);
    MFS_INHAUDIT(mth);

    /* Do the name lookup */

    if ((error = LOOKUP_FOR_IOCTL(pn,
                                  UIO_USERSPACE,
                                  lookflag,
                                  extra_lookflag,
                                  NULL,
                                  &vp,
                                  cd)) != 0)
    {
        ASSERT(vp == NULL);
        goto out;
    }

    if (MFS_VPISMFS(vp)) {
        switch (VTOM(vp)->mn_hdr.mclass) {
          case MFS_LOOPCLAS:
            xvp = MVFS_CVP_TO_VP(MFS_CLRVP(vp));
            VN_HOLD(xvp);
            break;
          case MFS_VOBRTCLAS:
            if (flags & MVFS_IL_FLAG_NB_LOOKUP) {
                /* We're being asked to do a "no bind" lookup. At this point,
                 * vp contains the unbound VOB root so skip binding. Note that
                 * we want to leave the unbound root in vp here and not xvp.
                 */
                break;
            }

            /* In this case, we need to put the unbound vob root into xvp
            * and the bound vob root into vp. Note, vp must be NULL'ed here
            * in case there is an error.
            */
            xvp = vp;
            vp = NULL;

            if (vpp == NULL) {
                /* If the caller has only requested the "underneath" vnode skip
                 * the bindroot. At this point, the unbound vob root is in xvp
                 * and vp is NULL.
                 */
                break;
            }

            MVFS_CD_SET_CRED(ncd,
                             MVFS_VIEW_CREDS(xvp, MVFS_CD2CRED(cd), FALSE));

            /* This call returns the same vnode (not held an extra time) on
            ** error (i.e., vp = xvp), so watch out for that below (the cleanup
            ** code assumes that vp and xvp are different, or at least are held
            ** separately).
            */
            vp = mfs_bindroot(xvp, ncd, &error);

            MVFS_CD_UNSET_CRED(ncd);

            if (error) {
                if (error == ESRCH) {
                    /* mfs_bindroot returns this error if there is no view
                     * context. Since were supposed to return with xvpp set to
                     * NULL for all cases other than VOBCLAS and LOOPCLAS,
                     * switch vp and xvp here (we moved the unbound root to xvp
                     * before the call to bindroot).
                     */
                    vp = xvp;
                    xvp = NULL;

                    /* Reset error here to avoid causing things to fail later
                     * on.
                     */
                    error = 0;
                } else {
                    /* Prevent a double rele in the cleanup code.  It's not
                    ** clear why mfs_bindroot returns vp set on error (see
                    ** above), but fixing it here for this case seems safest.
                    */
                    vp = NULL;
                    goto out;
                }
            }

            break;
          case MFS_VOBCLAS:
            /* We'll need to rebind the vnode after auditing is re-enabled which
             * might change the cleartext associated with it.
             */
          default:
            break;
        }
    } else {
        MVFS_VP_TO_CVP(vp, &cvp);

        if (MVOP_REALCVP(cvp, &rcvp) == 0  && rcvp != cvp) {
            /*
             * got a realvp through the operation, use that instead.
             */
            VN_HOLD(rcvp);
            CVN_RELE(cvp, cd);
            cvp = NULL;
            ATRIA_VN_RELE(vp, cd);
            vp = MVFS_CVP_TO_VP(rcvp);
            VN_HOLD(vp);
            CVN_RELE(rcvp, cd);
            rcvp = NULL;
        }
    }

out:
    /* Enable auditing */
    MFS_ENBAUDIT(mth);

    if (error != 0) {
        goto cleanup;
    }

    /* Rebind (if needed) when the vnode is in the VOB before the ioctl gets any
     * further, but after the audit inhibit is off. (rebind_vpp is inhibited by
     * the audit inhibit bit).  If there is an error, then vp should be null and
     * the VPISMFS macro will return false. Note here that vp may be NULL. This
     * happens in the case where we've looked up a vob root but the caller has
     * only asked for the "underneath" vnode (xvp).
     */
    if (vp != NULL && MFS_VPISMFS(vp) && MFS_ISVOB(VTOM(vp))) {
        MVFS_CD_SET_CRED(ncd, MVFS_VIEW_CREDS(vp, MVFS_CD2CRED(cd), FALSE));
        mfs_rebind_vpp(1, &vp, ncd);

        /* If caller is requesting a cleartext vnode (xvpp is not NULL) then we
         * need to make sure we fill it in. Note here that vp may be a bound vob
         * root (xvp is already filled in above) or a directory (no cleartext).
         * So, in these cases, skip this step.
         */
        if (xvpp != NULL && xvp == NULL && MVFS_ISVTYPE(vp, VREG)) {
            if ((error = mfs_getcleartext(vp, &xvp, cd)) != 0) {
                MVFS_CD_UNSET_CRED(ncd);
                goto cleanup;
            }
        }

        MVFS_CD_UNSET_CRED(ncd);
    }

    if (xvpp != NULL) {
        *xvpp = xvp;
        xvp = NULL;
    }

    if (vpp != NULL) {
        *vpp = vp;
        vp = NULL;
    }

cleanup:
    if (ncd != NULL) {
        MVFS_FREE_SUBSTITUTE_CRED(ncd);
    }

    if (vp != NULL) {
        ATRIA_VN_RELE(vp, cd);
    }

    if (cvp != NULL) {
        CVN_RELE(cvp, cd);
    }

    if (rcvp != NULL) {
        CVN_RELE(rcvp, cd);
    }

    if (xvp != NULL) {
        ATRIA_VN_RELE(xvp, cd);
    }

    return error;
}

/*
 * mvfs_ioctl_copyout
 *   This routine updates the status in the mvfs's cmdblock.
 *   NOTE: The user is now responsible for copying out data
 *         specific to the ioctl cmd.
 */
int
mvfs_ioctl_copyout(
    mvfscmd_block_t *data,
    mvfscmd_block_t *iocbufp,
    caddr_t infop,
    MVFS_CALLER_INFO *callinfo
)
{
    int error = 0;

    /* now get the status to be returned */

    error = CopyOutMvfscmd_block_status(iocbufp, (caddr_t) data, callinfo);
    return error;
}

/*
 * Other internal routines used to copyout info
 * for VOBINFO, VIEWINFO type calls.
 */

STATIC int MVFS_NOINLINE
mvfs_copyout_viewinfo(
    mvfs_viewinfo_t *vip,
    VNODE_T *vp
)
{
    VNODE_T *vw;
    int error = 0;
    mfs_mnode_t *mnp;

    vw = MFS_VIEW(vp);
    if (vw == NULL) {
        return EINVAL;
    }

    mnp = VTOM(vw);
    ASSERT(MFS_ISVIEW(mnp));

    MLOCK(mnp); /* Lock view object */

    /* Check for view w/out tagname */

    if (mnp->mn_view.viewname == NULL) {
        mvfs_log(MFS_LOG_ESTALE,
                 "ioctl: stale view vw=%s vob=%s dbid=0x%x\n",
                 mfs_vp2vw(vp),
                 mfs_vp2dev(vp),
                 mfs_vp2dbid(vp));
        MUNLOCK(mnp);
        return ESTALE;
    }

    error = mfs_copyout_strbuf(vip->vname, mnp->mn_view.viewname);
    if (error) {
        goto out;
    }

    error = mfs_copyout_strbufpn(
                vip->spath, MFS_STRBUFPN_PAIR_GET_UPN(&mnp->mn_view.svr.lpn).s);
    if (error) {
        goto out;
    }

    error = mfs_copyout_strbuf(vip->host, mnp->mn_view.svr.host);
    if (error) {
        goto out;
    }

    error = mfs_copyout_strbufpn(vip->rpath, mnp->mn_view.svr.rpn);
    if (error) {
        goto out;
    }

    vip->uuid = mnp->mn_view.svr.uuid;

out:
    MUNLOCK(mnp);

    return error;
}

STATIC int MVFS_NOINLINE
mvfs_copyout_vobinfo(
    mvfs_vobinfo_t *vip,
    struct mfs_mntinfo *mmi,
    tbs_boolean_t unique,
    mvfscmd_block_t *cmdblk,
    MVFS_CALLER_INFO *callinfo
)
{
    int error;

    error = mfs_copyout_strbufpn(vip->mntpath, mmi->mmi_mntpath);
    if (error) {
        return error;
    }

    error = mfs_copyout_strbufpn(
                    vip->spath, MFS_STRBUFPN_PAIR_GET_UPN(&mmi->mmi_svr.lpn).s);
    if (error) {
        return error;
    }

    error = mfs_copyout_strbuf(vip->host, mmi->mmi_svr.host);
    if (error) {
        return error;
    }

    error = mfs_copyout_strbufpn(vip->rpath, mmi->mmi_svr.rpn);
    if (error) {
        return error;
    }

    vip->oid = mmi->mmi_voboid;
    BCOPY(&mmi->mmi_vobuuid, &vip->uuid, sizeof(tbs_uuid_t));
    vip->unique = unique;
    error = CopyOutMvfs_vobinfo(vip, cmdblk->infop, callinfo);

    return error;
}

STATIC int MVFS_NOINLINE
mvfs_do_inval(
    mvfs_viewroot_type_t vrtype,
    tbs_oid_t *replica_oid_p,
    mvfs_ioinval_t *invp,
    VNODE_T *cdir,
    VNODE_T *vp,
    VFS_T *vfsp,
    tbs_boolean_t inval_local_only,
    CALL_DATA_T *cd
)
{
    int error = 0;
    struct alloc_unit {
        char *inv_nm;
        int mnum;
        VNODE_T *xvp;
        tbs_uuid_t vob_uuid;
        mfs_mnode_t *mnp;
        tbs_boolean_t safe;
        VFS_T *vfsp;
        CALL_DATA_T *ncd;
        CALL_DATA_T *nxcd;
        mvfs_vfs_data_t *mvdp;
    } *alloc_unitp = NULL;

    if ((alloc_unitp = KMEM_ALLOC(sizeof(*alloc_unitp), KM_SLEEP)) == NULL) {
        return ENOMEM;
    }

    BZERO(alloc_unitp, sizeof(*alloc_unitp));
    alloc_unitp->mvdp = MDKI_VFS_GET_DATAP(vp);

    ASSERT(MFS_VPISMFS(vp) && MFS_ISVIEW(VTOM(vp)));

    /* Now, switch on invalidate type.  Always invalidate
       the view first, then local information.  That way
       there should be no races that re-fetch stale info
       from the view (by another process) after the local
       object is invalidated, but before the view processes
       the invalidate request.  */
    alloc_unitp->ncd =
        MVFS_ALLOC_SUBSTITUTE_CRED(cd, MVFS_VIEW_CREDS(vp,
                                                       MVFS_CD2CRED(cd),
                                                       FALSE));

    if (!MVFS_SUBSTITUTE_CRED_IS_VALID(alloc_unitp->ncd)) {
        error = ENOMEM;
        goto cleanup;
    }

    alloc_unitp->nxcd = MVFS_ALLOC_SUBSTITUTE_CRED(cd, NULL);

    if (!MVFS_SUBSTITUTE_CRED_IS_VALID(alloc_unitp->nxcd)) {
        error = ENOMEM;
        goto cleanup;
    }

    switch (invp->invaltype) {
      case MVFS_INV_NC:
        error = mfs_copyin_strbufpn(invp->nm, &alloc_unitp->inv_nm);

        if (error) {
            break;
        }

        if (!inval_local_only) {
            error = mvfs_clnt_inval(vp,
                                    vfsp,
                                    VIEW_INVALIDATE_TYPE_NAME,
                                    replica_oid_p,
                                    &invp->obj_oid,
                                    alloc_unitp->inv_nm,
                                    alloc_unitp->ncd);
        }

        alloc_unitp->mnum = 0;
        alloc_unitp->vfsp = NULL;
        alloc_unitp->safe = FALSE;

        while ((alloc_unitp->mnp = mfs_mngetnextoid(&alloc_unitp->mnum,
                                                    vp,
                                                    replica_oid_p,
                                                    &invp->obj_oid)) != NULL)
        {
            /*
             * On Linux MVFS_VNGET will return ENOENT if the vnode is in the
             * process of going away and leave it to the caller to unwind or
             * wait, as needed.
             */
            if ((error = MVFS_VNGET(alloc_unitp->mnp->mn_hdr.vfsp,
                                    NULL,
                                    alloc_unitp->mnp,
                                    &alloc_unitp->xvp,
                                    cd)) != 0)
            {
                continue;
            }

            MLOCK(VTOM(alloc_unitp->xvp));

            /* Before DNLC so that rddir cache does not lag */
            MFS_ATTRINVAL(alloc_unitp->xvp);

            /* remove name */
            MVFS_CD_SET_CRED(alloc_unitp->nxcd,
                             MVFS_VIEW_CREDS(alloc_unitp->xvp,
                                             MVFS_CD2CRED(cd),
                                             TRUE));

            alloc_unitp->safe = (mfs_dncremove(alloc_unitp->xvp,
                                               alloc_unitp->inv_nm,
                                               alloc_unitp->nxcd) != 0);
            /*
             * If we were able to find the entry in the DNC, then
             * the RVC was purged if appropriate and safe == TRUE.
             */
            /* take copy of last known vfsp */
            alloc_unitp->vfsp = alloc_unitp->mnp->mn_hdr.vfsp;
            MUNLOCK(VTOM(alloc_unitp->xvp));
            ATRIA_VN_RELE(alloc_unitp->xvp, alloc_unitp->nxcd);
            alloc_unitp->xvp = NULL;
            MVFS_CD_UNSET_CRED(alloc_unitp->nxcd);
        }

        STRFREE(alloc_unitp->inv_nm);
        /*
         * We need to flush this replica's VOB root for this view,
         * because the RVC entry probably was not known under the name
         * that was just flushed, or the mnode might have been purged
         * from the cache before we got here and thus wasn't found by
         * mngetnextoid().  In either case the DNC RVC entry may still
         * be marked valid yet have the wrong fid.
         */
        if (!alloc_unitp->safe) {
            if (!alloc_unitp->vfsp) {
                MVFS_LOCK(&(alloc_unitp->mvdp->mount_lock));
                alloc_unitp->vfsp =
                    mvfs_findvfs_uuid(alloc_unitp->mvdp,
                                      &replica_oid_p->obj_uuid);
                MVFS_UNLOCK(&(alloc_unitp->mvdp->mount_lock));
                /*
                 * We don't really care if the VFS_T pointed to by vfsp is
                 * GC'ed--we don't reference the structure's contents.  We
                 * only care that any DNC entries which would point to
                 * mnodes in it get flushed.  (That will happen if the VOB
                 * is unmounted, so if the VFS goes and another one is
                 * allocated later, all we suffer is potentially excess
                 * cache flushing, not correctness problems.)
                 */
            }

            mvfs_rvcflush(vp, alloc_unitp->vfsp);
        }

        error = 0;
        break;
      case MVFS_INV_ROLEMAP:
        if (!inval_local_only) {
            error = mvfs_clnt_inval(vp,
                                    vfsp,
                                    VIEW_INVALIDATE_TYPE_ROLEMAP,
                                    replica_oid_p,
                                    &invp->obj_oid,
                                    NULL,
                                    alloc_unitp->ncd);
        }

        mvfs_acl_inval_oid(VOB_MTYPE_ROLEMAP, &invp->obj_oid);
        break;
      case MVFS_INV_POLICY:
        if (!inval_local_only) {
            error = mvfs_clnt_inval(vp,
                                    vfsp,
                                    VIEW_INVALIDATE_TYPE_POLICY,
                                    replica_oid_p,
                                    &invp->obj_oid,
                                    NULL,
                                    alloc_unitp->ncd);
        }

        mvfs_acl_inval_oid(VOB_MTYPE_POLICY, &invp->obj_oid);
        break;
      case MVFS_INV_OBJ:
      case MVFS_INV_ELEM:
        if (!inval_local_only) {
            error = mvfs_clnt_inval(vp,
                                    vfsp,
                                    VIEW_INVALIDATE_TYPE_OBJ,
                                    replica_oid_p,
                                    &invp->obj_oid,
                                    NULL,
                                    alloc_unitp->ncd);
        }

        alloc_unitp->mnum = 0;
        alloc_unitp->safe = FALSE;

        while ((alloc_unitp->mnp = mfs_mngetnextoid(&alloc_unitp->mnum,
                                                    vp,
                                                    replica_oid_p,
                                                    &invp->obj_oid)) != NULL)
        {
            /*
             * On Linux MVFS_VNGET will return ENOENT if the vnode is in the
             * process of going away and leave it to the caller to unwind or
             * wait, as needed.
             */
            if ((error = MVFS_VNGET(alloc_unitp->mnp->mn_hdr.vfsp,
                                    NULL,
                                    alloc_unitp->mnp,
                                    &alloc_unitp->xvp,
                                    cd)) != 0)
            {
                continue;
            }

            MLOCK(VTOM(alloc_unitp->xvp));

            MVFS_CD_SET_CRED(alloc_unitp->nxcd,
                             MVFS_VIEW_CREDS(alloc_unitp->xvp,
                                             MVFS_CD2CRED(cd),
                                             TRUE));

            MFS_ATTRINVAL(alloc_unitp->xvp);

            if (MVFS_ISVTYPE(alloc_unitp->xvp, VREG)) {
                mfs_clear_mark_purge(alloc_unitp->xvp);

                /*
                 * Don't leave the cleartext null, as some routines
                 * (especially paging routines) have trouble fetching
                 * it.  If we can't get it...  well, too bad, not much
                 * we can do here.  Leave it null and let the paging
                 * routine get the error.
                 */
            }

            if (!alloc_unitp->safe && invp->invaltype == MVFS_INV_ELEM) {
                /*
                 * Update the dir attribute cache flush time, so all of
                 * the ac and rdc entries before this time will be flushed.
                 * It's needed by "rmelem" operation only.
                 */
                VFS_TO_MMI(alloc_unitp->mnp->mn_hdr.vfsp)->mmi_ac_dir_ftime =
                    MDKI_CTIME();
            }

            /*
             * If the object being invalidated happens to be the bound VOB
             * root for this view, this call to mfs_dnc_invalvp() will flush
             * an RVC entry from the DNC, so set a flag to avoid doing it
             * again.
             */
            alloc_unitp->safe = TRUE;

            /*
             * Invalidate name cache for translations to and from
             * this object.
             */
            mfs_dnc_invalvp(alloc_unitp->xvp);
            MUNLOCK(VTOM(alloc_unitp->xvp));
            ATRIA_VN_RELE(alloc_unitp->xvp, alloc_unitp->nxcd);
            MVFS_CD_UNSET_CRED(alloc_unitp->nxcd);
            alloc_unitp->xvp = NULL;
        }

        /*
         * Must flush all ENOENTS that might change without any
         * change to the dir (e.g. label change could make object
         * appear).  See the name cache code for more info on this
         * call.
         */
        mfs_dnc_inval_obj_not_found();

        if (!alloc_unitp->safe) {
            /* Since we only get here if we didn't find any vnodes,
             * there was no sense in saving the vfsp.
             */
            MVFS_LOCK(&(alloc_unitp->mvdp->mount_lock));
            alloc_unitp->vfsp = mvfs_findvfs_uuid(alloc_unitp->mvdp,
                                                  &replica_oid_p->obj_uuid);
            MVFS_UNLOCK(&(alloc_unitp->mvdp->mount_lock));

            if (alloc_unitp->vfsp) {
                if (invp->invaltype == MVFS_INV_ELEM)
                    VFS_TO_MMI(alloc_unitp->vfsp)->mmi_ac_dir_ftime =
                        MDKI_CTIME();

                /*
                 * Must remove this VOB root for this view from the
                 * DNC, because flushing by mnode didn't find one.  If
                 * there isn't a mount for this replica, don't flush
                 * since there's nothing to worry about.
                 */
                mvfs_rvcflush(vp, alloc_unitp->vfsp);
            }
        }

        error = 0;
        break;
      case MVFS_INV_VIEW:
        if (!inval_local_only) {
            error = mvfs_clnt_inval(vp,
                                    vfsp,
                                    VIEW_INVALIDATE_TYPE_VIEW,
                                    &TBS_OID_NULL,
                                    &TBS_OID_NULL,
                                    NULL,
                                    alloc_unitp->ncd);
        }

        /* What about active vnodes?  HACK: Just do cdir! */
        if (MFS_VPISMFS(cdir) && MFS_ISVOB(VTOM(cdir))) {
            /* Before DNLC so that rddir cache does not lag */
            MFS_ATTRINVAL(cdir);
        }

        /* mvfs_dnc_invalvw() effects include an rvcflush */
        mfs_dnc_invalvw(vp); /* Invalidate name cache for this view */

        mfs_mnflushvw(vp, cd); /* Flush cached vnodes for view */
        break;
      case MVFS_INV_VFS:
        /* Invalidate VFS invalidates the whole mount point. */
        if (!inval_local_only) {
            error = mvfs_clnt_inval(vp,
                                    vfsp,
                                    VIEW_INVALIDATE_TYPE_VOB,
                                    replica_oid_p,
                                    &TBS_OID_NULL,
                                    NULL,
                                    alloc_unitp->ncd);
        }

        /* HACK: invalidate attrs on u_cdir if mfs */
        if (MFS_VPISMFS(cdir) && MFS_ISVOB(VTOM(cdir))) {
            /* Before DNLC so that rddir cache does not lag */
            MFS_ATTRINVAL(cdir);
        }

        mvfs_dncflush(vrtype, cd); /* Flush whole name cache */
        mvfs_mnflush(vrtype, cd);  /* Flush all vnodes */
        /* RVC was flushed as part of flush of name cache */
        break;
      default:
        error = EINVAL;
        break;
    }

cleanup:
    if (alloc_unitp->ncd != NULL) {
        MVFS_FREE_SUBSTITUTE_CRED(alloc_unitp->ncd);
    }

    if (alloc_unitp->nxcd != NULL) {
        MVFS_FREE_SUBSTITUTE_CRED(alloc_unitp->nxcd);
    }

    KMEM_FREE(alloc_unitp, sizeof(*alloc_unitp));

    return error;
}

/* Dispatch table for MVFS ioctls
 * NOTE: The ordering in this table must match the defines in mfs_ioctl.h.
 */
static mvfs_ioctl_subr_t mvfs_ioctl_disp[MVFS_CMD_MAX] = {
    mvfs_ioctl_get_viewinfo, /* MVFS_CMD_GET_VIEWINFO (1) */
    mvfs_ioctl_xstat, /* MVFS_CMD_XSTAT (2) */
    mvfs_ioctl_get_clrname, /* MVFS_CMD_GET_CLRNAME (3) */
    mvfs_ioctl_get_vfh, /* MVFS_CMD_GET_VFH (4) */
    mvfs_ioctl_ioinval, /* MVFS_CMD_IOINVAL (5) */
    mvfs_ioctl_audit_ioctl, /* MVFS_CMD_REVALIDATE (6) */
    mvfs_ioctl_mkviewtag, /* MVFS_CMD_MKVIEWTAG (7) */
    mvfs_ioctl_rmviewtag, /* MVFS_CMD_RMVIEWTAG (8) */
    mvfs_ioctl_get_viewaddr, /* MVFS_CMD_GET_VIEWADDR (9) */
    mvfs_ioctl_get_vobinfo, /* MVFS_CMD_GET_VOBINFO (10) */
    mvfs_ioctl_get_viewtag_dir, /* MVFS_CMD_GET_VIEWTAG_DIR (11) */
    mvfs_ioctl_change_mtype, /* MVFS_CMD_CHANGE_MTYPE (12) */
    mvfs_ioctl_audit_ioctl, /* MVFS_CMD_GET_AFILE (13) */
    mvfs_ioctl_audit_ioctl, /* MVFS_CMD_SET_AFILE (14) */
    mvfs_ioctl_audit_ioctl, /* MVFS_CMD_GET_PROCF (15) */
    mvfs_ioctl_audit_ioctl, /* MVFS_CMD_SET_PROCF (16) */
    mvfs_ioctl_audit_ioctl, /* MVFS_CMD_START_AUDIT (17) */
    mvfs_ioctl_audit_ioctl, /* MVFS_CMD_STOP_AUDIT (18) */
    mvfs_ioctl_audit_ioctl, /* MVFS_CMD_SYNC_AUDIT (19) */
    mvfs_ioctl_get_vxsuffix, /* MVFS_CMD_GET_VXSUFFIX (20) */
    mvfs_ioctl_get_cache_enb, /* MVFS_CMD_GET_CACHE_ENB (21) */
    mvfs_ioctl_set_cache_enb, /* MVFS_CMD_SET_CACHE_ENB (22) */
    mvfs_ioctl_flush_cache, /* MVFS_CMD_FLUSH_CACHE (23) */
    mvfs_ioctl_read_dnc, /* MVFS_CMD_READ_DNC (24) */
    mvfs_ioctl_get_rcsid, /* MVFS_CMD_GET_RCSID (25) */
    mvfs_ioctl_get_sccsid, /* MVFS_CMD_GET_SCCSID (26) */
    mvfs_ioctl_get_loginfo, /* MVFS_CMD_GET_LOGINFO (27) */
    mvfs_ioctl_set_loginfo, /* MVFS_CMD_SET_LOGINFO (28) */
    mvfs_ioctl_audit_ioctl, /* MVFS_CMD_GET_BH (29) */
    mvfs_ioctl_audit_ioctl, /* MVFS_CMD_SET_BH (30) */
    mvfs_ioctl_get_stats, /* MVFS_CMD_GET_STATS (31) */
    mvfs_ioctl_set_procview, /* MVFS_CMD_SETPROCVIEW (32) */
    mvfs_ioctl_get_procviewinfo, /* MVFS_CMD_GET_PROCVIEWINFO (33) */
    mvfs_ioctl_get_xattr, /* MVFS_CMD_GET_XATTR (34) */
    mvfs_ioctl_set_xattr, /* MVFS_CMD_SET_XATTR (35) */
    mvfs_ioctl_mount, /* MVFS_CMD_MOUNT (36) */
    mvfs_ioctl_unmount, /* MVFS_CMD_UNMOUNT (37) */
    mvfs_ioctl_rmallviewtags, /* MVFS_CMD_RMALLVIEWTAGS (38) */
    mvfs_ioctl_unmountall, /* MVFS_CMD_UNMOUNTALL (39) */
    mvfs_ioctl_get_poolmaps, /* MVFS_CMD_GET_POOLMAPS (40) */
    mvfs_ioctl_unsupported, /* MVFS_CMD_VDM (41) - NOT USED */
    mvfs_ioctl_abort, /* MVFS_CMD_ABORT (42) */
    mvfs_ioctl_exportviewtag, /* MVFS_CMD_EXPORTVIEWTAG (43) */
    mvfs_ioctl_unexportviewtag, /* MVFS_CMD_UNEXPORTVIEWTAG (44) */
    mvfs_ioctl_zero_stats, /* MVFS_CMD_ZERO_STATS (45) */
    mvfs_ioctl_get_cache_usage, /* MVFS_CMD_GET_CACHE_USAGE (46) */
    mvfs_ioctl_set_cache_sizes, /* MVFS_CMD_SET_CACHE_SIZES (47) */
    mvfs_ioctl_audit_ioctl, /* MVFS_CMD_AUDIT_MARKER (48) */
    mvfs_ioctl_unsupported, /* MVFS_CMD_IOD_NULL (49) - NOT USED */
    mvfs_ioctl_get_view_stats, /* MVFS_CMD_GET_VIEW_STATS (50) */
    mvfs_ioctl_zero_view_stats, /* MVFS_CMD_ZERO_VIEW_STATS (51) */
    mvfs_ioctl_sidhost_credmapping, /* MVFS_CMD_SIDHOST_CREDMAPPING (52) */
    mvfs_ioctl_delete_sidhost_credmapping, /* MVFS_CMD_DELETE_SIDHOST_CREDMAPPING (53) */
    mvfs_ioctl_get_viewtag_export, /* MVFS_CMD_GET_VIEWTAG_EXPORT (54) */
    mvfs_ioctl_get_gfsinfo, /* MVFS_CMD_GET_GFSINFO (55) */
    mvfs_ioctl_set_vobrt_vfsmnt, /* MVFS_CMD_SET_VOBRT_VFSMNT (56) */
    mvfs_ioctl_get_cache_sizes, /* MVFS_CMD_GET_CACHE_SIZES (57) */
    mvfs_ioctl_unsupported, /* 58 */
    mvfs_ioctl_unsupported, /* 59 */
    mvfs_ioctl_compute_cache_defaults, /* MVFS_CMD_COMPUTE_CACHE_DEFAULTS (60) */
    mvfs_ioctl_unsupported, /* 61 */
    mvfs_ioctl_unsupported, /* 62 */
    mvfs_ioctl_enable_pview_stats, /* MVFS_CMD_ENABLE_PVIEW_STATS (63) */
    mvfs_ioctl_disable_pview_stats /* MVFS_CMD_DISABLE_PVIEW_STATS (64) */
};

int
mvfs_mioctl(
    mvfs_viewroot_type_t vrtype,
    mvfscmd_block_t *cmdblk,
    CALL_DATA_T *cd,
    MVFS_CALLER_INFO *callinfo
)
{
    mvfs_ioctl_subr_t misp;
    int error = 0;

    /* MVFS_CMD_MIN is defined in sys/mfs_ioctl.h as '1' so we don't need to
     * worry about underflowing the table below.
     */
    if (MCB_CMD(cmdblk) > MVFS_CMD_MAX ||
        MCB_CMD(cmdblk) < MVFS_CMD_MIN)
    {
        return EINVAL;
    }

    /* The MVFS ioctls start at 1, but the ioctl dispatch starts at 0 so
     * subtract 1.
     */
    misp = mvfs_ioctl_disp[MCB_CMD(cmdblk) - 1];
    cmdblk->status = TBS_ST_OK;

    error =  misp(vrtype, cmdblk, cd, callinfo);

    if (error != 0 && cmdblk->status == TBS_ST_OK) {
        cmdblk->status = tbs_errno2status(error);
    }

    return error;
}

MVFS_IOCTL_HANDLER(get_viewinfo)
{
    int  error = 0;
    mvfs_viewinfo_t *viewinfo = NULL;
    VNODE_T *setview = NULL;
    VNODE_T *vp = NULL;

    if ((viewinfo = (mvfs_viewinfo_t *)KMEM_ALLOC(sizeof(*viewinfo),
                                                  KM_SLEEP)) == NULL)
    {
        return ENOMEM;
    }

    if ((error = mvfs_ioctl_lookup(vrtype,
                                   MCB_OBJPN(data),
                                   mvfs_get_ilflags_from_cb(data),
                                   &vp,
                                   NULL,
                                   cd,
                                   callinfo)) != 0)
    {
        goto cleanup;
    }

    ASSERT(vp != NULL);

    if ((error = CopyInMvfs_viewinfo(data->infop,
                                     viewinfo,
                                     callinfo)) != 0)
    {
        goto cleanup;
    }

    setview = mfs_getview(NULL, MVFS_CD2CRED(cd), TRUE /* HOLD */);

    if (!MFS_VPISMFS(vp) ||
        MFS_VIEW(vp) == NULL ||
        (MFS_VIEW(vp) == setview && MFS_ISLOOP(VTOM(vp))))
    {
        /* Don't return view info for loopback items in "our" view.
         * Some kernels have to cover every non-MVFS object, even the
         * ones in the current view.  Refusing view info for them
         * allows the auditing checks in clearmake et.al. to pass.
         * LOOKUP_FOR_AUDIT() in the MVFS auditing code will ensure
         * that the caller doesn't accidentally pass in an MVFS path
         * for an auditing file.
         */
        data->status = TBS_ST_NOT_AN_OBJ;
    } else {
        error = mvfs_copyout_viewinfo(viewinfo, vp);

        if (error == 0) {
            error = CopyOutMvfs_viewinfo(viewinfo, data->infop, callinfo);
        }
    }

cleanup:
    if (viewinfo != NULL) {
        KMEM_FREE(viewinfo, sizeof(*viewinfo));
    }

    if (vp != NULL) {
        ATRIA_VN_RELE(vp, cd);
    }

    if (setview != NULL) {
        ATRIA_VN_RELE(setview, cd);
    }

    return error;
}

MVFS_IOCTL_HANDLER(xstat)
{
    VATTR_T *vap = NULL;
    view_vstat_t *vstatp = NULL;
    CALL_DATA_T *ncd = NULL;
    int  error = 0;
    VNODE_T *vp = NULL;
    /* Use alloc_unitp to save stack space. */
    struct {
        mvfs_xstat_t xstat;
        mfs_mnode_t *mnp;
    } *alloc_unitp;

    if ((alloc_unitp = KMEM_ALLOC(sizeof(*alloc_unitp), KM_SLEEP)) == NULL) {
        return(ENOMEM);
    }
    BZERO(alloc_unitp, sizeof(*alloc_unitp));

    /* For xstat, we don't care about what's "underneath" the returned vnode.
     */
    if ((error = mvfs_ioctl_lookup(vrtype,
                                   MCB_OBJPN(data),
                                   mvfs_get_ilflags_from_cb(data),
                                   &vp,
                                   NULL,
                                   cd,
                                   callinfo)) != 0)
    {
        goto cleanup;
    }

    /*
     * Check for user buffer too small
     */
#if !(defined(ATRIA_LP64) || defined(ATRIA_LLP64))
    if (data->infolen < sizeof(alloc_unitp->xstat)) {
        error = EINVAL;
        goto cleanup;
    }
#endif

    if ((error = CopyInMvfs_xstat(data->infop, 
                                  &(alloc_unitp->xstat), 
                                  callinfo)) != 0)
    {
        goto cleanup;
    }

    vstatp = (view_vstat_t *)KMEM_ALLOC(sizeof(*vstatp), KM_SLEEP);
    if (vstatp == NULL) {
        error = ENOMEM;
        goto cleanup;
    }

    vap = MVFS_VATTR_ALLOC();
    if (vap == NULL) {
        error = ENOMEM;
        goto cleanup;
    }

    VATTR_NULL(vap);
    VATTR_SET_MASK(vap, AT_ALL);

    /*
     * Use VOP_GETATTR to get stats for 2 reasons:
     *   - It works on non-MFS objects
     *   - For MFS objects it does all the hokum about
     *     kludging size, devs etc. for the external world.
     */
    ncd = MVFS_ALLOC_SUBSTITUTE_CRED(cd,
                                     MVFS_VIEW_CREDS(vp,
                                                     MVFS_CD2CRED(cd),
                                                     FALSE));

    if (MVFS_SUBSTITUTE_CRED_IS_VALID(ncd)) {
        if (MFS_VPISMFS(vp)) {
            error = mfs_getattr(vp, vap, 0, ncd);
        } else {
            /* No ports use the cvp field of MVOP_GETATTR, so, save a lookup
             * here and pass NULL.
             */
            error = MVOP_GETATTR(vp, NULL, vap, 0, ncd);
        }
    } else {
        error = ENOMEM;
    }

    if (error != 0) {
        goto cleanup;
    }

    /*
     * Now, create initial vstatp for copyout.
     * Create this from the VATTR_T record,
     * and add in any oids as needed.
     * CCM28830: size is now ks_off64_t, we should no longer cast to u_long
     */
    BZERO(vstatp, sizeof(*vstatp));
    switch(VATTR_GET_TYPE(vap)) {
        case VREG: vstatp->fstat.type = TBS_FTYPE_REG; break;
        case VDIR: vstatp->fstat.type = TBS_FTYPE_DIR; break;
        case VBLK: vstatp->fstat.type = TBS_FTYPE_BLK; break;
        case VCHR: vstatp->fstat.type = TBS_FTYPE_CHR; break;
        case VLNK: vstatp->fstat.type = TBS_FTYPE_LNK; break;
        case VFIFO: vstatp->fstat.type = TBS_FTYPE_FIFO; break;
        default: vstatp->fstat.type = TBS_FTYPE_NULL; break;
    }

    vstatp->fstat.mode =
        (tbs_fmode_t)(VATTR_GET_MODE(vap) & VIEW_ATTR_FMODES_MASK);
    vstatp->fstat.nlink = VATTR_GET_NLINK(vap);
    MVFS_VATTR_TO_FSTAT_DB_UID(vap, &vstatp->fstat.usid);
    MVFS_VATTR_TO_FSTAT_DB_GID(vap, &vstatp->fstat.gsid);
    vstatp->fstat.size = VATTR_GET_SIZE(vap);
    vstatp->fstat.nodeid = VATTR_GET_NODEID(vap);
    VATTR_GET_ATIME_TV(vap, &vstatp->fstat.atime);
    VATTR_GET_MTIME_TV(vap, &vstatp->fstat.mtime);
    VATTR_GET_CTIME_TV(vap, &vstatp->fstat.ctime);

    /* Set oids/mtype only if MFS object under VOB
     */
    if (MFS_VPISMFS(vp)) {
        alloc_unitp->mnp = VTOM(vp);

        (alloc_unitp->xstat).flags = 0;

        switch ((alloc_unitp->mnp)->mn_hdr.mclass) {
          case MFS_VOBCLAS:
            /* Copy over extended stat fields */
            vstatp->mtype = (alloc_unitp->mnp)->mn_vob.attr.mtype;
            vstatp->elem_oid = (alloc_unitp->mnp)->mn_vob.attr.elem_oid;
            vstatp->obj_oid  = (alloc_unitp->mnp)->mn_vob.attr.obj_oid;
            vstatp->event_time = (alloc_unitp->mnp)->mn_vob.attr.event_time;
            vstatp->rolemap_oid  = (alloc_unitp->mnp)->mn_vob.attr.rolemap_oid;
            vstatp->eacl_mtime = (alloc_unitp->mnp)->mn_vob.attr.eacl_mtime;
            (alloc_unitp->xstat).vob_oid = V_TO_MMI(vp)->mmi_voboid;
            (alloc_unitp->xstat).replica_uuid = V_TO_MMI(vp)->mmi_vobuuid;

            if (MFS_VIEW(vp)) {
                (alloc_unitp->xstat).view_uuid = 
                    VTOM(MFS_VIEW(vp))->mn_view.svr.uuid;
                (alloc_unitp->xstat).view_hm = VTOM(MFS_VIEW(vp))->mn_view.hm;
            } else {
                (alloc_unitp->xstat).view_uuid = TBS_UUID_NULL;
                (alloc_unitp->xstat).view_hm   = 0;
            }

            (alloc_unitp->xstat).xmode = 
                (alloc_unitp->mnp)->mn_vob.attr.fstat.mode & 
                ~VIEW_ATTR_FMODES_MASK;

            /*
             * Add in appropriate flags.
             */
            if (MFS_ISROOTSYNONYM(vp, V_TO_MMI(vp)->mmi_root_edbid)) {
                (alloc_unitp->xstat).flags |= MVFS_XISROOT;
            }

            if (VTOM(MFS_VIEW(vp))->mn_view.hm) {
                (alloc_unitp->xstat).flags |= MVFS_XVXOBJ;
            }

            if (MFS_FSTAT_AUDITED(VTOM(vp))) {
                (alloc_unitp->xstat).flags |= MVFS_XAUDITED;
            }

            break;
          case MFS_VOBRTCLAS:
            /* User space uses this flag to detect if it xstat'ed an unbound
             * VOB root. We used to return TBS_ST_NOT_AN_OBJ here but that
             * really isn't the correct thing to do since this function "should
             * work for all objects including non-mvfs ones".
             */
            (alloc_unitp->xstat).flags |= MVFS_XISUROOT;
            break;
          default:
            /* Nothing special to do here. */
            break;
        }
    }

    if ((alloc_unitp->xstat).vstat != NULL &&
        (error = CopyOutView_vstat(vstatp,
                                   (caddr_t)(alloc_unitp->xstat).vstat,
                                   callinfo)) != 0)
    {
        goto cleanup;
    }

    error = CopyOutMvfs_xstat(&(alloc_unitp->xstat), data->infop, callinfo);

cleanup:
    if (vp != NULL) {
        ATRIA_VN_RELE(vp, cd);
    }

    if (vap != NULL) {
        MVFS_FREE_VATTR_FIELDS(vap);
        MVFS_VATTR_FREE(vap);
    }

    if (vstatp != NULL) {
        KMEM_FREE(vstatp, sizeof(*vstatp));
    }

    if (ncd != NULL) {
        MVFS_FREE_SUBSTITUTE_CRED(ncd);
    }
    KMEM_FREE(alloc_unitp, sizeof(*alloc_unitp));

    return error;
}

MVFS_IOCTL_HANDLER(get_clrname)
{
    mvfs_clrname_info_t clrname_info;
    mfs_mnode_t *mnp;
    int  error = 0;
    VNODE_T *vp = NULL;

    if ((error = mvfs_ioctl_lookup(vrtype,
                                   MCB_OBJPN(data),
                                   mvfs_get_ilflags_from_cb(data),
                                   &vp,
                                   NULL,
                                   cd,
                                   callinfo)) != 0)
    {
        goto cleanup;
    }

    if ((error = CopyInMvfs_clrname_info(data->infop,
                                         &clrname_info,
                                         callinfo)) != 0)
    {
        goto cleanup;
    }

    if (!MFS_VPISMFS(vp)) {
        error = EINVAL;
        goto cleanup;
    }

    mnp = VTOM(vp);
    if (!MFS_ISVOB(mnp)) { /* Only on vob objects for now */
        error = EINVAL;
        goto cleanup;
    }

    MLOCK(mnp);

    /*
     * Get cleartext name only if caller has asked for it.
     * This causes extra RPC's and may result in a "fetch"
     * of the cleartext, so it should be looked upon as an
     * "expensive" operation.
     */
    if (clrname_info.clrname.s != NULL &&
        clrname_info.clrname.m != 0)
    {
        if (MVFS_ISVTYPE(vp, VREG) &&
            mnp->mn_vob.cleartext.nm == NULL)
        {
            /* XXX Was MVFS_VIEW_CREDS(vp, cd, TRUE) Is this the same as the
             * passed in call_data?
             */
            error = mfs_getcleartext(vp, NULL, cd);
        } else {
            /* No cleartext expected, or already have it. */
            error = 0;
        }

        if (mnp->mn_vob.cleartext.nm && error == 0) {
            if (mnp->mn_vob.cleartext.isvob) {
                clrname_info.clrpool = MVFS_XCTVOB;
            } else {
                clrname_info.clrpool = MVFS_XCTVIEW;
            }

            error = mfs_copyout_strbufpn(clrname_info.clrname,
                                         mnp->mn_vob.cleartext.nm);
        } else {
            if (error) {
                clrname_info.clrpool = MVFS_XCTERROR;
            } else {
                clrname_info.clrpool = MVFS_XCTNONE;
            }

            /*
             * Overwrite error on copy-out, we are done with it
             */
            error = mfs_copyout_strbufpn(clrname_info.clrname, "");
        }

        error = CopyOutMvfs_clrname_info(&clrname_info,
                                         data->infop,
                                         callinfo);
    }

    MUNLOCK(mnp);

cleanup:
    if (vp != NULL) {
        ATRIA_VN_RELE(vp, cd);
    }

    return error;
}

MVFS_IOCTL_HANDLER(get_vfh)
{
    int  error = 0;
    mfs_mnode_t *mnp;
    mvfs_iovfh_t iovfh;
    VNODE_T *vp = NULL;

    if ((error = mvfs_ioctl_lookup(vrtype,
                                   MCB_OBJPN(data),
                                   mvfs_get_ilflags_from_cb(data),
                                   &vp,
                                   NULL,
                                   cd,
                                   callinfo)) != 0)
    {
        goto cleanup;
    }

    if (!MFS_VPISMFS(vp)) {
        error = EINVAL;
        goto cleanup;
    }

    if ((error = CopyInMvfs_iovfh(data->infop, &iovfh, callinfo)) == 0) {
        mnp = VTOM(vp);

        if (!MFS_ISVOB(mnp)) { /* Only on vob objects for now */
            error = EINVAL;
        } else {
            iovfh.vfh = mnp->mn_vob.vfh; /* set view handle */
            error = CopyOutMvfs_iovfh(&iovfh, data->infop, callinfo);
        }
    }

cleanup:
    if (vp != NULL) {
        ATRIA_VN_RELE(vp, cd);
    }

    return error;
}

MVFS_IOCTL_HANDLER(ioinval)
{
    int  error = 0;
    mvfs_ioinval_t *ioinvalp = NULL;
    /* Use alloc_unitp to save stack space. */
    struct {
        VFS_T *vobvfs;
        mfs_mnode_t *mnp;
    } *alloc_unitp;
    tbs_uuid_t vob_uuid;
    tbs_oid_t *vob_oid_p;
    tbs_boolean_t unique;
    int indx, first;
    mvfs_vfs_data_t *mvdp = NULL;
    VNODE_T *vp = NULL;
    VNODE_T *cdir = NULL;
    tbs_boolean_t inval_local_only = FALSE;

    cdir = MDKI_GET_U_CDIR();

    if ((alloc_unitp = KMEM_ALLOC(sizeof(*alloc_unitp), KM_SLEEP)) == NULL) {
        return(ENOMEM);
    }
    if (cdir == NULL) {
        error = MVFS_NULL_U_CDIR_ERR;
        goto cleanup;
    }

    /* Use vp as a temporary variable here to save stack space. */
    if (MVOP_REALVP(cdir, &vp) == 0 && cdir != vp) {
        /*
         * there was an underlying real VP, we'd better use that.
         */
        MDKI_VNHOLD_RCDIR(vp);
        MDKI_VNRELE_RCDIR(cdir);
        cdir = vp; /* no need to frob ref cnts */
    }

    /* Make sure vp is NULL for the call to mvfs_ioctl_lookup. See the comment
     * on ref counts above regarding the absense of a rele.
     */
    vp = NULL;

    if ((error = mvfs_ioctl_lookup(vrtype,
                                   MCB_OBJPN(data),
                                   mvfs_get_ilflags_from_cb(data),
                                   &vp,
                                   NULL,
                                   cd,
                                   callinfo)) != 0)
    {
        goto cleanup;
    }

    if (!MFS_VPISMFS(vp)) {
        error = EINVAL;
        goto cleanup;
    }

    if ((ioinvalp = KMEM_ALLOC(sizeof(*ioinvalp), KM_SLEEP)) == NULL) {
        error = ENOMEM;
        goto cleanup;
    }

    alloc_unitp->mnp = VTOM(vp);

    if ((error = CopyInMvfs_ioinval(data->infop, ioinvalp, callinfo)) != 0) {
        goto cleanup;
    }

    switch (ioinvalp->utype) {
      case MVFS_IOINVAL_VOB_OID_LOCAL_ONLY:
        inval_local_only = TRUE;
        /* Fall through */
      case MVFS_IOINVAL_VOB_OID:
        if (!MFS_ISVIEW(alloc_unitp->mnp)) {
            error = EINVAL;
            goto cleanup;
        }

        /*
         * Locate vob uuid
         * We may have multiple replicas of the same VOB
         * family mounted.  Loop through the vobvfs table
         * invalidating all the ones we find.
         *
         * We ignore races of new things added to the
         * MVFS mountlist while we do the invalidations.
         */
        error = indx = 0;
        first = 1;
        mvdp = MDKI_VFS_GET_DATAP(vp);
        do {
            MVFS_LOCK(&(mvdp->mount_lock));
            alloc_unitp->vobvfs =
                mvfs_findvfs_oid(mvdp,
                                 &ioinvalp->un.vob_oid,
                                 &indx,
                                 &unique);
            MVFS_UNLOCK(&(mvdp->mount_lock));
            if (alloc_unitp->vobvfs == NULL) {
                if (first) {
                    error = EINVAL;
                }

                break;
            }

            vob_uuid = VFS_TO_MMI(alloc_unitp->vobvfs)->mmi_vobuuid;
            vob_oid_p = (tbs_oid_t *)&vob_uuid;
            error = mvfs_do_inval(vrtype,
                                  vob_oid_p,
                                  ioinvalp,
                                  cdir,
                                  vp,
                                  alloc_unitp->vobvfs,
                                  inval_local_only,
                                  cd);
            first = 0;
            if (error)
                break;
        } while (alloc_unitp->vobvfs != NULL);

        break;
      case MVFS_IOINVAL_REPLICA_UUID_LOCAL_ONLY:
        inval_local_only = TRUE;
        /* Fall through */
      case MVFS_IOINVAL_REPLICA_UUID:
        if (!MFS_ISVIEW(alloc_unitp->mnp)) {
            error = EINVAL;
            goto cleanup;
        }

        error = mvfs_do_inval(vrtype,
                              (tbs_oid_t *)&ioinvalp->un.replica_uuid,
                              ioinvalp,
                              cdir,
                              vp,
                              vp->v_vfsp,
                              inval_local_only,
                              cd);
        break;
      case MVFS_IOINVAL_ATTRCACHE_PN:
        /* In this case, vp is the vnode corresponding to the path which
         * we need to invalidate the attributes for.  Note, that no locks
         * are needed here since MFS_ATTRINVAL only involves a one word
         * write.
         */
        if (MFS_ISVOB(alloc_unitp->mnp)) {
            MFS_ATTRINVAL(vp);
        } else {
            error = EINVAL;
        }

        break;
      default:
        error = EINVAL;
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_mioctl MVFS_CMD_IOINVAL: Unknown invalidation "
                 "target (%u)\n",
                 ioinvalp->utype);
        break;
    }

  cleanup:
    if (ioinvalp != NULL) {
        KMEM_FREE(ioinvalp, sizeof(*ioinvalp));
    }

    if (vp != NULL) {
        ATRIA_VN_RELE(vp, cd);
    }

    if (cdir != NULL) {
        MDKI_VNRELE_RCDIR(cdir);
    }
    KMEM_FREE(alloc_unitp, sizeof(*alloc_unitp));

    return error;
}

MVFS_IOCTL_HANDLER(audit_ioctl)
{
    return mfs_auditioctl(data, cd, callinfo);
}

MVFS_IOCTL_HANDLER(mkviewtag)
{
    mvfs_mkviewtag_info_t *vtp = NULL;
    mvfs_mkviewtag_info_ex_t *vtp_ex = NULL;
    int error = 0;

    /* We have two mkviewtag ioctls (although one is Windows only for
    ** now).  They use two structs, but the mkviewtag one is a subset of
    ** the mkviewtag_ex one (the net_path field is added), so make them
    ** common here so the mvfs_mkviewtag code can be common as well.
    */
    if ((vtp_ex = KMEM_ALLOC(sizeof(*vtp_ex), KM_SLEEP)) == NULL) {
        error = ENOMEM;
        goto cleanup;
    }

    if ((vtp = KMEM_ALLOC(sizeof(*vtp), KM_SLEEP)) == NULL) {
        error = ENOMEM;
        goto cleanup;
    }

    if ((error = CopyInMvfs_mkviewtag_info(data->infop, vtp, callinfo)) != 0)
    {
        goto cleanup;
    }

    vtp_ex->viewtag = vtp->viewtag;
    vtp_ex->spath = vtp->spath;
    vtp_ex->host = vtp->host;
    vtp_ex->rpath = vtp->rpath;
    MFS_INIT_STRBUFPN(&(vtp_ex->net_path)); /* The added field. */
    vtp_ex->uuid = vtp->uuid;
    vtp_ex->addr = vtp->addr;
    vtp_ex->windows_view = vtp->windows_view;
    vtp_ex->pad = vtp->pad;

    error = mvfs_mkviewtag_subr(vrtype, data, vtp_ex, cd, callinfo);

cleanup:
    if (vtp != NULL) {
        KMEM_FREE(vtp, sizeof(*vtp));
    }

    if (vtp_ex != NULL) {
        KMEM_FREE(vtp_ex, sizeof(*vtp_ex));
    }

    return error;
}

STATIC int MVFS_NOINLINE
mvfs_mkviewtag_subr(
    mvfs_viewroot_type_t vrtype,
    mvfscmd_block_t *data,
    mvfs_mkviewtag_info_ex_t *vtp_ex,
    CALL_DATA_T *cd,
    MVFS_CALLER_INFO *callinfo
)
{
    int  error = 0;
    VNODE_T *vw = NULL;
    VNODE_T *dvp = NULL;
    char *rpn = NULL;
    char *net_pn = NULL;
    char *host = NULL;
    char *tagn = NULL;
    CRED_T *vwcreds = NULL;
    tbs_boolean_t spath_copied_in = FALSE; /* Needed for cleanup. */
    mfs_mnode_t *mnp;
    /* Declare a type so we can do one allocation to save stack space. */
    struct {
        struct mfs_svr svr;
        tbs_uuid_str_t vw_uuid_str;
    } *alloc_unitp = NULL;
    struct mfs_svr *svrp;
    tbs_uuid_str_t *vw_uuid_strp = NULL;
    mvfs_viewroot_data_t *vrdp = MDKI_VIEWROOT_GET_DATAP(vrtype);
    tbs_status_t status = TBS_ST_OK;

    /* Allocate stuff to keep the stack size down. */
    if ((alloc_unitp = KMEM_ALLOC(sizeof(*alloc_unitp), KM_SLEEP)) == NULL) {
        return ENOMEM;
    }

    svrp = &(alloc_unitp->svr);
    vw_uuid_strp = &(alloc_unitp->vw_uuid_str);

    data->status = TBS_ST_OK;

    if (vrtype == DYVWROOT) {
        /* Check to see if the new addr that was passed in via the ioctl can
         * actually be contacted before assuming that it is valid.  Use a
         * contact RPC to the view server at that address to do so.  That way we
         * can also learn the server version so we can tell what functions it
         * supports. Note, only need to do this for dynamic views, since
         * automatic views use the lpc mechanism.
         */
        BZERO(svrp, sizeof(*svrp));
        svrp->addr = vtp_ex->addr;
        svrp->down = 0;
        svrp->svrbound = 1;
        if ((error = mvfs_clnt_ping_server(vrtype,
                                           svrp,
                                           MVFS_CD2CRED(cd))) != 0)
        {
            mvfs_log(MFS_LOG_DEBUG,
                     "mvfs_mkviewtag: mvfs_clnt_ping_server vrtype=%d"
                     " failed with error=%d\n",
                     vrtype, error);
            goto cleanup;
        }
    }

    /* Get parameters */
    if ((error = mfs_copyin_strbuf(vtp_ex->viewtag, &tagn)) != 0) {
        goto cleanup;
    }

    /* We must prevent the history mode suffix being part of the view name. */
    if (mvfs_hmname_type(vrtype, tagn, NULL)) {
        error = EINVAL;
        goto cleanup;
    }

    /* We are going to use the pathname strings, so copy them in.  The
    ** CopyInMvfs_mkviewtag_info_ex above already did the equivalent of
    ** CopyInMfs_strbufpn_pair since mkviewtag_info_ex.spath is a
    ** mfs_strbufpn_pair_t.
    */
    MFS_STRBUFPN_PAIR_COPYIN_STRS(&(vtp_ex->spath), error);
    if (error != 0) {
        goto cleanup;
    }

    spath_copied_in = TRUE;

    /* Get the view server hostname, remote view storage pathname, and maybe the
    ** network pathname.  Differentiate between the mkviewtag and mkviewtag_ex
    ** ioctls based on whether or not the net_path is filled in.
    **
    ** XXXODFS: I'm not sure we really need any of these in the context of
    ** automatic views but a preliminary glance shows they're used all over the
    ** place for logging, so, for now, fill them in.
    */
    if (((error = mfs_copyin_strbuf(vtp_ex->host, &host)) == 0) &&
        ((error = mfs_copyin_strbufpn(vtp_ex->rpath, &rpn)) == 0) &&
        (vtp_ex->net_path.s == NULL ||
         (error = mfs_copyin_strbufpn(vtp_ex->net_path, &net_pn)) == 0))
    {
        /* We need to prevent an add/add race, as that will result in user-space
         * code getting EEXIST and believing the directory needs to be
         * removed/recreated, which in a thundering herd problem results in
         * spurious errors.
         */
        MVFS_LOCK(&(vrdp->mvfs_mkviewtag_lock));

        /* See if existing view-tag */
        if ((error = mvfs_viewtaglookup(vrtype, tagn, &vw, cd)) == 0) {
            mnp = VTOM(vw);
            ASSERT(MFS_ISVIEW(mnp));
            MLOCK(mnp);
            /*
             * See if all setup OK already
             * This compare is always case-sensitive
             */
            if (PN_STRCMP(FALSE,
                          MFS_STRBUFPN_PAIR_GET_KPN(&(mnp->mn_view.svr.lpn)).s,
                          MFS_STRBUFPN_PAIR_GET_KPN(&(vtp_ex->spath)).s) == 0 &&
                HN_STRCMP(mnp->mn_view.svr.host, host) == 0 &&
                PN_STRCMP(FALSE, mnp->mn_view.svr.rpn, rpn) == 0 &&
                MFS_UUIDEQ(mnp->mn_view.svr.uuid, vtp_ex->uuid))
            {
                /* Always fix with latest address */
                mnp->mn_view.svr.addr = vtp_ex->addr;
                mnp->mn_view.svr.svrbound = 1;  /* Reset flags */
                mnp->mn_view.svr.dprinted = 0;
                mnp->mn_view.svr.uprinted = 0;
                data->status = TBS_ST_OK;
            } else {
                /*
                 * Exists, but not same info, give EEXIST error
                 */
                data->status = TBS_ST_EEXIST;
            }

            MUNLOCK(mnp);
            mvfs_rvcflush(vw, NULL); /* Flush root version cache */
            if (vw) {
                ATRIA_VN_RELE(vw, cd);
            }
        } else {
            /* Look to see if some process might still be attached to
             * it (but the tag was removed by another process). If so,
             * resurrect it.
             */
            error = mvfs_viewuuidrecover(tagn,
                                         &(vtp_ex->uuid),
                                         host,
                                         &(vtp_ex->spath),
                                         rpn,
                                         &vw,
                                         cd);
            switch (error) {
              case 0:
                mnp = VTOM(vw);
                MLOCK(mnp);
                mnp->mn_view.svr.addr = vtp_ex->addr;
                mnp->mn_view.svr.svrbound = 1;  /* Reset flags */
                mnp->mn_view.svr.dprinted = 0;
                mnp->mn_view.svr.uprinted = 0;
                data->status = TBS_ST_OK;
                MUNLOCK(mnp);
                mvfs_rvcflush(vw, NULL); /* Flush root version cache */
                ATRIA_VN_RELE(vw, cd);
                goto out;

              default:
                /* some problem reattaching; return this error */
                data->status = tbs_errno2status(error);
                error = 0;
                goto out;

              case ENOENT:
                /* not found to reattach, let's do the hard work */
                break;
            }
            /* Doesn't exist... try to create it */
            dvp = MVFS_GETVIEWROOT(vrtype);
            if (dvp == NULL) {
                data->status = TBS_ST_MFS_ERR;
                error = 0;
            }

            /* Make the view-tag.  The windows_view will be FALSE (from the
            ** caller) for Unix views and set correctly for Windows views, so we
            ** can just use it in either case.
            */
            else if ((error = mfs_viewdirmkdir(dvp,
                                               tagn,
                                               NULL,
                                               &vw,
                                               cd,
                                               host,
                                               vtp_ex->windows_view)) != 0)
            {
                data->status = tbs_errno2status(error);
                error = 0;
            }
            /* We made the view tag so finish everything off. */
            else {
                ASSERT(vw != NULL);
                mnp = VTOM(vw);
                ASSERT(MFS_ISVIEW(mnp));
                /*
                 * Move string ptrs into viewtag.  Null out the local ptrs so
                 * they are not freed in caller.
                 */
                MLOCK(mnp); /* Lock while updating */

                /* windows_view was filled in during mfs_viewdirmkdir call, but
                ** maybe it could change...
                */
                mnp->mn_view.windows_view = vtp_ex->windows_view;
                mnp->mn_view.auto_view = (vrtype == AUVWROOT ? 1 : 0);
                mnp->mn_view.svr.lpn = vtp_ex->spath;

                /* Don't use MFS_STRBUFPN_PAIR_GET_{KPN,UPN} macros because we
                ** want to make sure we are setting both fields to NULL so we
                ** don't free them during cleanup (since we copied them).
                */
                vtp_ex->spath.kpn.s = NULL;
                vtp_ex->spath.upn.s = NULL;

                /* View server hostname.  Also, set ptrs to NULL after they're
                ** copied so we don't free them at the end during cleanup.
                */
                mnp->mn_view.svr.rpn = rpn; rpn = NULL;

                mnp->mn_view.svr.net_pn = net_pn; net_pn = NULL;
                mnp->mn_view.svr.uuid = vtp_ex->uuid;
                mnp->mn_view.svr.addr = vtp_ex->addr;
                mnp->mn_view.svr.svrbound = 1;  /* reset flags */
                mnp->mn_view.svr.dprinted = 0;
                mnp->mn_view.svr.uprinted = 0;
                /*
                 * Make view handle be UUID too.  This catches cases where I am
                 * talking to a different view than I think I am.
                 */
                mnp->mn_view.vh.view_uuid = vtp_ex->uuid;

                vwcreds = MDKI_GET_UCRED();
                mnp->mn_view.cuid = MDKI_CR_GET_UID(vwcreds);
                mnp->mn_view.cgid = MDKI_CR_GET_GID(vwcreds);
                MDKI_CRFREE(vwcreds);

                switch (vrtype) {
                  default:
                    mvfs_log(MFS_LOG_ERR,
                             "mvfs_mkviewtag: Unknown viewroot type %d, "
                             "assuming dynamic viewroot\n",
                             vrtype);
                    /* Fall through */
                  case DYVWROOT:
                    mnp->mn_view.vob_mnts = NULL;
                    MVFS_AUVW_VOBTAG_DIRP(mnp) = NULL;
                    break;
                  case AUVWROOT:
                    if (mnp->mn_view.vob_mnts != NULL) {
                        /* We don't currently recycle view mnodes, but, just in
                         * case we do at some point in the future clean and
                         * reinitialize the vfs data structure so that it gets
                         * the correct tag and has coherent information.
                         */
                        mvfs_vfs_data_clean(mnp->mn_view.vob_mnts);
                    } else {
                        /* Set up this views VOB mount structure */
                        mnp->mn_view.vob_mnts =
                           KMEM_ALLOC(sizeof(*mnp->mn_view.vob_mnts), KM_SLEEP);

                        if (mnp->mn_view.vob_mnts == NULL) {
                            error = ENOMEM;
                            break;
                        } else {
                            BZERO(mnp->mn_view.vob_mnts,
                                  sizeof(*mnp->mn_view.vob_mnts));
                        }
                    }

                    error = mvfs_vfs_data_init(MVFS_VDIF_ALL,
                                               mnp->mn_view.vob_mnts,
                                               tagn);

                    /* Setup the vobtag dir structure for the (auto) view. */
                    if (MVFS_AUVW_VOBTAG_DIRP(mnp) != NULL) {

                        /* We don't currently recycle view mnodes, but, just in
                        ** case we do at some point in the future clean and
                        ** reinitialize the vfs data structure so that it gets
                        ** the correct tag and has coherent information.
                        */
                        mvfs_vobtag_dir_clean(mnp, cd);
                    }
                    error = mvfs_vobtag_dir_init(mnp);

                    /* Don't bother checking error here. If there was an error
                     * the VOB mount structure (and lpc handle and vobtag dir)
                     * we just allocated will get cleaned up below.
                     */
                    break;
                }

              unlock_mn:
                /* If there was an error clean up any per view vob mount
                 * structures or lpc handles we may have created while we still
                 * have the mnode lock.
                 */
                if (error != 0) {
                    if (mnp->mn_view.vob_mnts != NULL) {
                        mvfs_vfs_data_clean(mnp->mn_view.vob_mnts);
                        KMEM_FREE(mnp->mn_view.vob_mnts,
                                  sizeof(*mnp->mn_view.vob_mnts));
                        mnp->mn_view.vob_mnts = NULL;
                    }
                    if (MVFS_AUVW_VOBTAG_DIRP(mnp) != NULL) {
                        mvfs_vobtag_dir_clean(mnp, cd);
                    }
                }

                MUNLOCK(mnp);

                if (vw != NULL) {
                    mvfs_rvcflush(vw, NULL); /* Flush root version cache */
                    ATRIA_VN_RELE(vw, cd);
                }

                /* If there was an error, remove the view from the viewroot. Do
                 * this after we've unlocked the mnode and performed an RVC flush.
                 */
                if (error != 0) {
                    mfs_viewdirrmdir(dvp, tagn, cd);
                    data->status = tbs_errno2status(error);
                }
            }
        }

      out:
        MVFS_UNLOCK(&(vrdp->mvfs_mkviewtag_lock));
        if (dvp != NULL) {
            ATRIA_VN_RELE(dvp, cd);
        }
    }

  cleanup:
    if (net_pn != NULL) {
        PN_STRFREE(net_pn);
    }

    if (rpn != NULL) {
        PN_STRFREE(rpn);
    }

    if (host != NULL) {
        HN_STRFREE(host);
    }

    if (spath_copied_in) {
        PNPAIR_STRFREE(&(vtp_ex->spath));
    }

    if (tagn != NULL) {
        PN_STRFREE(tagn);
    }

    if (alloc_unitp != NULL) {
        KMEM_FREE(alloc_unitp, sizeof(*alloc_unitp));
    }

    return error;
}

MVFS_IOCTL_HANDLER(rmviewtag)
{
    int  error = 0;
    mvfs_viewtag_info_t viewtag_info;
    VNODE_T *dvp = NULL;
    char *tagn = NULL;

    if ((error = CopyInMvfs_viewtag_info(data->infop,
                                         &viewtag_info,
                                         callinfo)) == 0)
    {
        error = mfs_copyin_strbuf(viewtag_info.viewtag, &tagn);
        if (error == 0 ) {
            /* Get held vnode ptr to viewroot dir */
            dvp = MVFS_GETVIEWROOT(vrtype);
            if (dvp == NULL) {
                error = ENXIO;
            } else {
                /* Now, remove the view-tag */
                error = mfs_viewdirrmdir(dvp, tagn, cd);
                ATRIA_VN_RELE(dvp, cd);
            }

            PN_STRFREE(tagn);
        }
    }

    return error;
}

MVFS_IOCTL_HANDLER(get_viewaddr)
{
    int  error = 0;
    struct mvfs_viewaddr viewaddr;
    mfs_mnode_t *mnp;
    VNODE_T *vp = NULL;

    if ((error = mvfs_ioctl_lookup(vrtype,
                                   MCB_OBJPN(data),
                                   mvfs_get_ilflags_from_cb(data),
                                   &vp,
                                   NULL,
                                   cd,
                                   callinfo)) != 0)
    {
        return error;
    }

    if (!MFS_VPISMFS(vp) || MFS_VIEW(vp) == NULL) {
        error = EINVAL;
    } else if ((error = CopyInMvfs_viewaddr(data->infop,
                                            &viewaddr,
                                            callinfo)) == 0)
    {
        mnp = VTOM(MFS_VIEW(vp));
        ASSERT(MFS_ISVIEW(mnp));
        viewaddr.addr = mnp->mn_view.svr.addr;
        error = CopyOutMvfs_viewaddr(&viewaddr, data->infop, callinfo);
    }

    ATRIA_VN_RELE(vp, cd);

    return error;
}

MVFS_IOCTL_HANDLER(get_vobinfo)
{
    int  error = 0;
    struct mfs_mntinfo *mmi;
    VFS_T *vfsp;
    tbs_boolean_t unique;
    tbs_uuid_t uuid;
    tbs_oid_t oid;
    mvfs_vobinfo_t *vobinfop = NULL;
    VNODE_T *vp = NULL;
    mvfs_vfs_data_t *mvdp = NULL;
    mvfs_il_flags_t ilflags = MVFS_IL_FLAG_NONE;
    mfs_strbufpn_t pnp;

    if ((vobinfop = KMEM_ALLOC(sizeof(*vobinfop), KM_SLEEP)) == NULL) {
        return ENOMEM;
    }

    if ((error = CopyInMvfs_vobinfo(data->infop, vobinfop, callinfo)) != 0) {
        goto cleanup;
    }

    if (vobinfop->utype == MVFS_VOBINFO_IN_PNAME) {
        pnp = MFS_STRBUFPN_PAIR_GET_KPN(&(vobinfop->vobid.pn_s.pair));
        ilflags =
            mvfs_ioctl_lookup_pnflags_to_ilflags(vobinfop->vobid.pn_s.pnflags);

        /* Do a "no bind" lookup. In this case, mvfs_ioctl_lookup will return an
         * unbound VOB root without attempting to bind it. We don't want to
         * attempt a bindroot here because there are cases where we actually
         * have a view but are unable to properly execute a bindroot (e.g. no
         * version of the VOB root is selected by the view). It's also important
         * to note here that we cant just ask mvfs_ioctl_lookup for the
         * underlying node because this routine must work with all paths
         * including things underneath the root of a VOB (see mfs_ioctl.h). If
         * we're given a path to something underneath the root,
         * mvfs_ioctl_lookup would return the cleartext vnode.
         */
        ilflags |= MVFS_IL_FLAG_NB_LOOKUP;

        if ((error = mvfs_ioctl_lookup(vrtype,
                                       pnp,
                                       ilflags,
                                       &vp,
                                       NULL,
                                       cd,
                                       callinfo)) != 0)
        {
            goto cleanup;
        }
    } else {
        if (vrtype == AUVWROOT) {
            /* In the case of automatic views, there should be a viewtag in the
             * ioctl header. We need to lookup the view for this viewtag and get
             * the vfs data pointer for the view from there.
             */
            if ((error = mvfs_ioctl_lookup(vrtype,
                                           MCB_OBJPN(data),
                                           MVFS_IL_FLAG_PN_IS_VIEWTAG,
                                           &vp,
                                           NULL,
                                           cd,
                                           callinfo)) != 0)
            {
                goto cleanup;
            } else {
                if (!MFS_VPISMFS(vp) || !MFS_ISVIEW(VTOM(vp))) {
                    goto cleanup;
                }
            }
        }

        /* vp may be NULL in the case of DYVWROOT here, this is ok as
         * MDKI_VFS_GET_DATAP returns the Dynamic view vfs data pointer when
         * passed NULL.
         */
        mvdp = MDKI_VFS_GET_DATAP(vp);
    }

    switch (vobinfop->utype) {
      case MVFS_VOBINFO_IN_PNAME:
        ASSERT(vp);
        if (!MFS_VPISMFS(vp) ||
            (!MFS_ISVOB(VTOM(vp)) && !MFS_ISVOBRT(VTOM(vp))))
        {
            data->status = TBS_ST_NOT_AN_OBJ;
            break;
        }

        mmi = V_TO_MMI(vp); /* Get mount data */
        error = mvfs_copyout_vobinfo(vobinfop, mmi, TRUE, data, callinfo);
        break;
      case MVFS_VOBINFO_IN_UUID:
        error = CopyInTbs_uuid_s((caddr_t)vobinfop->vobid.vob_uuid, &uuid,
                                 callinfo);
        if (error) {
            break;
        }

        MVFS_LOCK(&(mvdp->mount_lock));
        vfsp = mvfs_findvfs_uuid(mvdp, &uuid);
        if (vfsp == NULL) {
            MVFS_UNLOCK(&(mvdp->mount_lock));
            data->status = TBS_ST_NOT_FOUND;
            break;
        }

        error = mvfs_copyout_vobinfo(vobinfop, VFS_TO_MMI(vfsp), TRUE, data,
                                     callinfo);
        MVFS_UNLOCK(&(mvdp->mount_lock));
        break;
      case MVFS_VOBINFO_IN_OID:
        error = CopyInTbs_oid_s((caddr_t)vobinfop->vobid.vob_oid, (&oid),
                                callinfo);
        if (error) {
            break;
        }

        MVFS_LOCK(&(mvdp->mount_lock));
        vfsp = mvfs_findvfs_oid(mvdp, &oid, NULL, &unique);
        if (vfsp == NULL) {
            MVFS_UNLOCK(&(mvdp->mount_lock));
            data->status = TBS_ST_NOT_FOUND;
            break;
        }

        error = mvfs_copyout_vobinfo(vobinfop, VFS_TO_MMI(vfsp), unique,
                                     data, callinfo);
        MVFS_UNLOCK(&(mvdp->mount_lock));
        break;
      case MVFS_VOBINFO_IN_COOKIE:
        /*
         * Find vfs by cookie and update to next entry to find.
         */
        MVFS_LOCK(&(mvdp->mount_lock));
        vfsp = mvfs_findvfs_cookie(mvdp, &(vobinfop->vobid.cookie));
        if (vfsp == NULL) {
            MVFS_UNLOCK(&(mvdp->mount_lock));
            data->status = TBS_ST_NOT_FOUND;
            break;
        }

        error = mvfs_copyout_vobinfo(vobinfop, VFS_TO_MMI(vfsp), TRUE, data,
                                     callinfo);
        MVFS_UNLOCK(&(mvdp->mount_lock));
        break;
      default:
          error = ENOTTY;
          break;
    }

cleanup:
    if (vp != NULL) {
        ATRIA_VN_RELE(vp, cd);
    }

    if (vobinfop != NULL) {
        KMEM_FREE(vobinfop, sizeof(*vobinfop));
    }

    return error;
}

MVFS_IOCTL_HANDLER(get_viewtag_dir)
{
    int  error = 0;
    mfs_strbufpn_pair_t get_viewtag_dir_info;
    mfs_strbufpn_t *strbufpnp;
    mvfs_viewroot_data_t *vrdp = MDKI_VIEWROOT_GET_DATAP(vrtype);
    /* Viewroots are mounted in the "global" vfs data structure. Since we're
     * being asked for the mount path of the viewroot corresponding to the given
     * mvfs_viewroot_type_t, use the global vfs data structure.
     */
    mvfs_vfs_data_t *mvdp = MDKI_VFS_GET_DATAP(NULL);

    if ((error = CopyInMfs_strbufpn_pair(data->infop,
                                         &get_viewtag_dir_info,
                                         callinfo)) == 0)
    {
        strbufpnp = &get_viewtag_dir_info.upn;
        MVFS_LOCK(&(mvdp->mount_lock));
        if (vrdp->mfs_viewroot_vfsp) {
            error = mfs_copyout_strbufpn(*strbufpnp,
                            VFS_TO_MMI(vrdp->mfs_viewroot_vfsp)->mmi_mntpath);
        } else {
            error = ENODEV;
        }

        if (!error) {
            strbufpnp = &get_viewtag_dir_info.kpn;
            if (vrdp->mfs_viewroot_vfsp) {
                error = mfs_copyout_strbufpn(*strbufpnp,
                    VFS_TO_MMI(vrdp->mfs_viewroot_vfsp)->mmi_mntpath);
            } else {
                error = ENODEV;
            }
        }

        MVFS_UNLOCK(&(mvdp->mount_lock));

        /*
         * Need to Copy the length information out to the user
         */
        error = CopyOutMfs_strbufpn_pair(&get_viewtag_dir_info,
                                         data->infop,
                                         callinfo);
    }

    return error;
}

MVFS_IOCTL_HANDLER(change_mtype)
{
    mvfs_iochange_mtype_t change_mtype;
    CALL_DATA_T *ncd = NULL;
    VNODE_T *vp = NULL;
    int error = 0;

    if ((error = CopyInMvfs_iochange_mtype(data->infop,
                                           &change_mtype,
                                           callinfo)) != 0)
    {
        goto cleanup;
    }

    if ((error = mvfs_ioctl_lookup(vrtype,
                                   MCB_OBJPN(data),
                                   mvfs_get_ilflags_from_cb(data),
                                   &vp,
                                   NULL,
                                   cd,
                                   callinfo)) != 0)
    {
        goto cleanup;
    }

    if (!MFS_VPISMFS(vp)) {
        error = EINVAL;
        goto cleanup;
    }

    ncd = MVFS_ALLOC_SUBSTITUTE_CRED(cd,
                                     MVFS_VIEW_CREDS(vp,
                                                     MVFS_CD2CRED(cd),
                                                     FALSE));

    if (!MVFS_SUBSTITUTE_CRED_IS_VALID(ncd)) {
        error = ENOMEM;
        goto cleanup;
    }

    error = mfs_clnt_change_mtype(vp,
                                  change_mtype.mtype,
                                  &data->status,
                                  ncd);

cleanup:
    if (ncd != NULL) {
        MVFS_FREE_SUBSTITUTE_CRED(ncd);
    }

    if (vp != NULL) {
        ATRIA_VN_RELE(vp, cd);
    }

    return error;
}

MVFS_IOCTL_HANDLER(get_vxsuffix)
{
    int error = 0;
    mvfs_viewroot_data_t *vrdp = MDKI_VIEWROOT_GET_DATAP(vrtype);
    mvfs_vfs_data_t *mvdp;
    mfs_strbuf_t vxsuffix_data;

    if ((error = CopyInMfs_strbuf(data->infop,
                                  &vxsuffix_data,
                                  callinfo)) != 0)
    {
        return error;
    }

    /* Since we're going to look at the viewroot, get the global vfs
     * data structure (indicated by the NULL arg).
     */
    mvdp = MDKI_VFS_GET_DATAP(NULL);

    MVFS_LOCK(&(mvdp->mount_lock));
    if (vrdp->mfs_viewroot_vfsp) {
        error = mfs_copyout_strbuf(
            vxsuffix_data, VFS_TO_MMI(vrdp->mfs_viewroot_vfsp)->mmi_hmsuffix);
    } else {
        error = ENODEV;
    }

    MVFS_UNLOCK(&(mvdp->mount_lock));

    error = CopyOutMfs_strbuf(&vxsuffix_data, data->infop, callinfo);

    return error;
}

MVFS_IOCTL_HANDLER(get_cache_enb)
{
    u_long caches_enabled = 0;
    mvfs_common_data_t *mcdp = MDKI_COMMON_GET_DATAP();

    if (mcdp->mvfs_acenabled) caches_enabled |= MVFS_CE_ATTR;
    if (mcdp->mvfs_dncenabled) caches_enabled |= MVFS_CE_NAME;
    if (mcdp->mvfs_dncnoentenabled) caches_enabled |= MVFS_CE_NOENT;
    if (mcdp->mvfs_rvcenabled) caches_enabled |= MVFS_CE_RVC;
    if (mcdp->mvfs_rlenabled) caches_enabled |= MVFS_CE_SLINK;
    if (mcdp->mvfs_rdcenabled) caches_enabled |= MVFS_CE_RDC;
    if (mcdp->mvfs_ctoenabled) caches_enabled |= MVFS_CE_CTO;
    if (mcdp->mvfs_rebind_dir_enable) caches_enabled |= MVFS_CE_CWDREBIND;

    return CopyOutMvfs_u_long(&caches_enabled, data->infop, callinfo);
}

MVFS_IOCTL_HANDLER(set_cache_enb)
{
    u_long caches_to_enable = 0;
    int error = 0;
    mvfs_common_data_t *mcdp = MDKI_COMMON_GET_DATAP();

    if (!MDKI_SUSER(MVFS_CD2CRED(cd))) {
        MDKI_SET_U_ERROR(0);   /* Eliminate suser side effect */
        return EPERM;
    }

    if ((error = CopyInMvfs_u_long(data->infop,
                                   &caches_to_enable,
                                   callinfo)) == 0)
    {
        mcdp->mvfs_acenabled = (caches_to_enable & MVFS_CE_ATTR) ? 1 : 0;
        mcdp->mvfs_dncenabled = (caches_to_enable & MVFS_CE_NAME) ? 1 : 0;
        mcdp->mvfs_dncnoentenabled = (caches_to_enable & MVFS_CE_NOENT) ? 1 : 0;
        mcdp->mvfs_rvcenabled = (caches_to_enable & MVFS_CE_RVC) ? 1 : 0;
        mcdp->mvfs_rlenabled = (caches_to_enable & MVFS_CE_SLINK) ? 1 : 0;
        mcdp->mvfs_rdcenabled = (caches_to_enable & MVFS_CE_RDC) ? 1 : 0;
        mcdp->mvfs_ctoenabled = (caches_to_enable & MVFS_CE_CTO) ? 1 : 0;
        mcdp->mvfs_rebind_dir_enable = ((caches_to_enable & MVFS_CE_CWDREBIND) ?
                                        1 :
                                        0);
    }

    return error;
}

MVFS_IOCTL_HANDLER(flush_cache)
{
    int error = 0;
    u_long cache_flush_flags = 0;

    if ((error = CopyInMvfs_u_long(data->infop,
                                   &cache_flush_flags,
                                   callinfo)) != 0)
    {
        return error;
    }

    if (cache_flush_flags & MVFS_CF_MN) {
        mvfs_mnflush(vrtype, cd);
    }

    if (cache_flush_flags & MVFS_CF_NC) {
        mvfs_dncflush(vrtype, cd);
    }

    if (cache_flush_flags & MVFS_CF_RVC) {
        mvfs_viewdirflushrvc(vrtype, cd);
    }

    if (cache_flush_flags & MVFS_CF_LKUP) {
        MVFS_FLUSH_CREDLIST(TRUE);
    }

    return error;
}

MVFS_IOCTL_HANDLER(read_dnc)
{
    int  error = 0;
    mfs_ioncent_t *ioncentp;

    if ((ioncentp = KMEM_ALLOC(sizeof(*ioncentp), KM_SLEEP)) == NULL) {
        return ENOMEM;
    }

    if ((error = CopyInMfs_ioncent(data->infop, ioncentp, callinfo)) == 0) {
        if ((error = mfs_dnc_getent(ioncentp, cd)) == 0) {
            error = CopyOutMfs_ioncent(ioncentp, data->infop, callinfo);
        }
    }

    KMEM_FREE(ioncentp, sizeof(*ioncentp));
    return error;
}

MVFS_IOCTL_HANDLER(get_rcsid)
{
    mfs_strbuf_t rcsid_data;
    int error = 0;

    if ((error = CopyInMfs_strbuf(data->infop,
                                  &rcsid_data,
                                  callinfo)) != 0)
    {
        return error;
    }

    if ((error = mfs_copyout_strbuf(rcsid_data, MVFS_RCSID_STRING_VAR)) != 0) {
        return error;
    }

    error = CopyOutMfs_strbuf(&rcsid_data, data->infop, callinfo);

    return error;
}

MVFS_IOCTL_HANDLER(get_sccsid)
{
    mfs_strbuf_t sccsid_data;
    int error = 0;

    if ((error = CopyInMfs_strbuf(data->infop,
                                  &sccsid_data,
                                  callinfo)) != 0)
    {
        return error;
    }

    if ((error = mfs_copyout_strbuf(sccsid_data,
                                    MVFS_SCCSID_STRING_VAR)) != 0)
    {
        return error;
    }

    error = CopyOutMfs_strbuf(&sccsid_data, data->infop, callinfo);

    return error;
}

MVFS_IOCTL_HANDLER(get_loginfo)
{
    int  error = 0;
    mvfs_loginfo_t loginfo;

    if ((error = CopyInMvfs_loginfo(data->infop, &loginfo, callinfo)) == 0)
    {
        loginfo.priority = mfs_logpri;
        loginfo.vops_mask = mdb_vops;
        loginfo.vfsops_mask = mdb_vfsops;
        loginfo.xops_mask = mdb_xops;
        loginfo.traps_mask = mdb_traps;
        loginfo.kernlog_pn.l = loginfo.kernlog_pn.m;
        loginfo.assert_panic_on = mvfs_panic_assert;
        if ((error = mvfs_logfile_get(loginfo.kernlog_pn.s,
                                      &loginfo.kernlog_pn.l)) == 0)
        {
            error = CopyOutMvfs_loginfo(&loginfo, data->infop, callinfo);
        }
    }

    return error;
}

MVFS_IOCTL_HANDLER(set_loginfo)
{
    int  error = 0;
    mvfs_loginfo_t loginfo;
    char *logfile = NULL;

    if ((error = CopyInMvfs_loginfo(data->infop, &loginfo, callinfo)) == 0)
    {

        if (!MDKI_SUSER(MVFS_CD2CRED(cd)) || !MDKI_INGLOBALZONE()) {
            error = EPERM;
            MDKI_SET_U_ERROR(0); /* Eliminate suser side effect.*/
        } else {
            if (loginfo.mask & MVFS_LOGINFO_KERNLOGFILE) {
                if (loginfo.kernlog_pn.l == 0) {
                    mvfs_logfile_close(cd);
                } else {
                    error = mfs_copyin_strbufpn(loginfo.kernlog_pn, &logfile);
                    if (error == 0) {
                        error = mvfs_logfile_set(logfile, cd);
                    }
                }
            }

            if (error == 0) {
                if (loginfo.mask & MVFS_LOGINFO_PRIORITY) {
                    mfs_logpri = loginfo.priority;
                    mvfs_mnclear_logbits();
                    MVFS_LOGBUFFER_FLUSH();
                }

                if (loginfo.mask & MVFS_LOGINFO_VOPS) {
                    mdb_vops = loginfo.vops_mask;
                }

                if (loginfo.mask & MVFS_LOGINFO_VFSOPS) {
                    mdb_vfsops = loginfo.vfsops_mask;
                }

                if (loginfo.mask & MVFS_LOGINFO_XOPS) {
                    mdb_xops = loginfo.xops_mask;
                }

                if (loginfo.mask & MVFS_LOGINFO_TRAPS) {
                    mdb_traps = loginfo.traps_mask;
                }

                if ((loginfo.mask & MVFS_LOGINFO_PANICONOFF) != 0) {
                    mvfs_panic_assert = loginfo.assert_panic_on;
                }

                error = 0;
            }

            if (logfile != NULL) {
                PN_STRFREE(logfile);
            }
        }
    }

    return error;
}

MVFS_IOCTL_HANDLER(get_stats)
{
    int  error = 0;
    int  cpuid;
    mvfs_statbufs_t *mvfs_statbufsp;
    mvfs_stats_data_t *percpu_sdp;
    mvfs_stats_data_t *output_sdp;

    /* If we have a CPU id beyond the max cpus, we print an error message before
     * returning the statistics.  We print a different error message for
     * platforms in which max cpu is obtained through an API and the ones which
     * define it as a constant value that is obtained from the header files.
     * This is because, in the former case, restarting ClearCase might help pick
     * up the right value.  MAXCPU_IS_CONSTANT is used to indicate if the max
     * cpu is constant.
     */
    if (mvfs_cpu_beyond_limit) {
        UPRINTF(("mvfs: ERROR: The current number of CPUs is larger than the "
                 "maximum number of CPUs(%d) MVFS could handle. Some "
                 "statistics might have been lost. \n", mvfs_max_cpus));
#ifndef MAXCPU_IS_CONSTANT
        UPRINTF(("Stopping and restarting ClearCase might fix this issue.\n"));
#endif
        UPRINTF(("Continuing with the available statistics....\n \n"));
    }

    if ((output_sdp = KMEM_ALLOC(sizeof(mvfs_stats_data_t), KM_SLEEP)) == NULL)
    {
        return ENOMEM;
    }

    MVFS_STAT_ZERO(output_sdp); /* This sets histvalrpc and histvallpc, too. */

    if ((mvfs_statbufsp = KMEM_ALLOC(sizeof(*mvfs_statbufsp),
                                     KM_SLEEP)) == NULL)
    {
        KMEM_FREE(output_sdp, sizeof(mvfs_stats_data_t));
        return ENOMEM;
    }

    for (cpuid = 0; cpuid < mvfs_max_cpus; cpuid++) {
        percpu_sdp = MDKI_STATS_GET_DATAP(cpuid);
        if ((percpu_sdp != NULL) && !(percpu_sdp->zero_me))  {
           mvfs_addup_clntstat(percpu_sdp, output_sdp);
           mvfs_addup_mnstat(percpu_sdp, output_sdp);
           mvfs_addup_clearstat(percpu_sdp, output_sdp);
           mvfs_addup_rvcstat(percpu_sdp, output_sdp);
           mvfs_addup_dncstat(percpu_sdp, output_sdp);
           mvfs_addup_acstat(percpu_sdp, output_sdp);
           mvfs_addup_rlstat(percpu_sdp, output_sdp);
           mvfs_addup_austat(percpu_sdp, output_sdp);
           mvfs_addup_eacstat(percpu_sdp, output_sdp);
           mvfs_addup_vnopcnt(percpu_sdp, output_sdp);
           mvfs_addup_vfsopcnt(percpu_sdp, output_sdp);
           mvfs_addup_viewopcnt(percpu_sdp, output_sdp);
           mvfs_addup_viewoptime(percpu_sdp, output_sdp);
           mvfs_addup_viewophist(percpu_sdp, output_sdp);
        }
    }

    if ((error = CopyInMvfs_statbufs(data->infop,
                                     mvfs_statbufsp, callinfo)) == 0)
    {

        if (mvfs_statbufsp->clntstat.s &&
            mvfs_statbufsp->clntstat.m)
        {
            error = CopyOutMfs_clntstat(&(output_sdp->mfs_clntstat),
                        (caddr_t)mvfs_statbufsp->clntstat.s,
                         mvfs_statbufsp->clntstat.m, callinfo);
        }

        if (error == 0 && mvfs_statbufsp->mnstat.s &&
            mvfs_statbufsp->mnstat.m)
        {
            error = COPYOUT((caddr_t)&(output_sdp->mfs_mnstat),
                            (caddr_t)mvfs_statbufsp->mnstat.s,
                            KS_MIN(mvfs_statbufsp->mnstat.m,
                                   sizeof(struct mfs_mnstat)));
        }

        if (error == 0 && mvfs_statbufsp->clearstat.s &&
            mvfs_statbufsp->clearstat.m)
        {
            error = CopyOutMfs_clearstat(&(output_sdp->mfs_clearstat),
                        (caddr_t)mvfs_statbufsp->clearstat.s,
                        mvfs_statbufsp->clearstat.m, callinfo);
        }

        if (error == 0 && mvfs_statbufsp->rvcstat.s &&
            mvfs_statbufsp->rvcstat.m)
        {
            error = COPYOUT((caddr_t)&(output_sdp->mfs_rvcstat),
                            (caddr_t)mvfs_statbufsp->rvcstat.s,
                            KS_MIN(mvfs_statbufsp->rvcstat.m,
                                   sizeof(struct mfs_rvcstat)));
        }

        if (error == 0  && mvfs_statbufsp->dncstat.s &&
            mvfs_statbufsp->dncstat.m)
        {
            error = COPYOUT((caddr_t)&(output_sdp->mfs_dncstat),
                            (caddr_t)mvfs_statbufsp->dncstat.s,
                            KS_MIN(mvfs_statbufsp->dncstat.m,
                                   sizeof(struct mfs_dncstat)));
        }

        if (error == 0 && mvfs_statbufsp->acstat.s &&
            mvfs_statbufsp->acstat.m)
        {
            error = COPYOUT((caddr_t)&(output_sdp->mfs_acstat),
                            (caddr_t)mvfs_statbufsp->acstat.s,
                            KS_MIN(mvfs_statbufsp->acstat.m,
                                   sizeof(struct mfs_acstat)));
        }

        if (error == 0 && mvfs_statbufsp->rlstat.s &&
            mvfs_statbufsp->rlstat.m)
        {
            error = COPYOUT((caddr_t)&(output_sdp->mfs_rlstat),
                            (caddr_t)mvfs_statbufsp->rlstat.s,
                            KS_MIN(mvfs_statbufsp->rlstat.m,
                                   sizeof(struct mfs_rlstat)));
        }

        if (error == 0 && mvfs_statbufsp->austat.s &&
            mvfs_statbufsp->austat.m)
        {
            error = CopyOutMfs_austat(&(output_sdp->mfs_austat),
                        (caddr_t)mvfs_statbufsp->austat.s,
                        mvfs_statbufsp->austat.m, callinfo);
        }

        if (error == 0 && mvfs_statbufsp->eacstat.s &&
            mvfs_statbufsp->eacstat.m)
        {
            error = COPYOUT((caddr_t)&(output_sdp->mvfs_eacstat),
                            (caddr_t)mvfs_statbufsp->eacstat.s,
                            KS_MIN(mvfs_statbufsp->eacstat.m,
                                   sizeof(struct mvfs_eacstat)));
        }

        /* Copy out the opcnt vectors */

        if (error == 0 && mvfs_statbufsp->vnopcnt.s &&
            mvfs_statbufsp->vnopcnt.m)
        {
            error = COPYOUT((caddr_t)output_sdp->mfs_vnopcnt,
                             (caddr_t)mvfs_statbufsp->vnopcnt.s,
                              KS_MIN(mvfs_statbufsp->vnopcnt.m,
                              sizeof(MVFS_STAT_CNT_T)*mfs_vnopmax));
        }

        if (error == 0 && mvfs_statbufsp->vfsopcnt.s &&
            mvfs_statbufsp->vfsopcnt.m)
        {
            error = COPYOUT((caddr_t)output_sdp->mfs_vfsopcnt,
                            (caddr_t)mvfs_statbufsp->vfsopcnt.s,
                             KS_MIN(mvfs_statbufsp->vfsopcnt.m,
                              sizeof(MVFS_STAT_CNT_T)*mfs_vfsopmax));
        }

        if (error == 0 && mvfs_statbufsp->viewopcnt.s &&
            mvfs_statbufsp->viewopcnt.m)
        {
            error = COPYOUT((caddr_t)output_sdp->mfs_viewopcnt,
                            (caddr_t)mvfs_statbufsp->viewopcnt.s,
                            KS_MIN(mvfs_statbufsp->viewopcnt.m,
                                   sizeof(MVFS_STAT_CNT_T) * mvfs_viewopmax));
        }

        if (error == 0 && mvfs_statbufsp->rviewopcnt.s &&
            mvfs_statbufsp->rviewopcnt.m)
        {
            error = COPYOUT((caddr_t)output_sdp->mvfs_rviewopcnt,
                            (caddr_t)mvfs_statbufsp->rviewopcnt.s,
                            KS_MIN(mvfs_statbufsp->rviewopcnt.m,
                                   sizeof(MVFS_STAT_CNT_T) * mvfs_viewopmax));
        }

        if (error == 0 && mvfs_statbufsp->viewoptime.s &&
            mvfs_statbufsp->viewoptime.m)
        {
            error = CopyOuttimestruc_array(
                output_sdp->mfs_viewoptime,
                (caddr_t)mvfs_statbufsp->viewoptime.s,
                mvfs_viewopmax,
                KS_MIN(mvfs_statbufsp->viewoptime.m,
                       sizeof(timestruc_t) * mvfs_viewopmax),
                callinfo);
        }

        if (error == 0 && mvfs_statbufsp->rviewoptime.s &&
            mvfs_statbufsp->rviewoptime.m)
        {
            error = CopyOuttimestruc_array(
                output_sdp->mvfs_rviewoptime,
                (caddr_t)mvfs_statbufsp->rviewoptime.s,
                mvfs_viewopmax,
                KS_MIN(mvfs_statbufsp->rviewoptime.m,
                       sizeof(timestruc_t) * mvfs_viewopmax),
                callinfo);
        }

        if (error == 0 && mvfs_statbufsp->viewophist.s &&
            mvfs_statbufsp->viewophist.m)
        {
            error = CopyOutMvfs_rpchist(&(output_sdp->mvfs_viewophist),
                                        (caddr_t)mvfs_statbufsp->viewophist.s,
                                        mvfs_statbufsp->viewophist.m,
                                        callinfo);
        }

    }

    KMEM_FREE(output_sdp, sizeof(mvfs_stats_data_t));
    KMEM_FREE(mvfs_statbufsp, sizeof(*mvfs_statbufsp));

    return error;
}

MVFS_IOCTL_HANDLER(set_procview)
{
    char *tagn;
    mvfs_viewtag_info_t viewtag_info;
    mfs_strbuf_t *strbuf;
    VNODE_T *vw;
    int  error = 0;

    if ((error = CopyInMvfs_viewtag_info(data->infop,
                                         &viewtag_info,
                                         callinfo)) != 0)
    {
        goto out;
    }

    strbuf = &viewtag_info.viewtag;

    /*
     * For this ioctl, allow a NULL string ptr.  This is how a user resets to
     * 'no current view'.  The copyin may return an error, but it always sets
     * tagn (NULL on error, the copied in string otherwise).  We don't care what
     * error is returned.
     */
    error = mfs_copyin_strbuf(*strbuf, &tagn);
    if (error == 0) {
     /* Look up new view tag to make sure it is OK */
        error = mvfs_viewtaglookup(vrtype, tagn, &vw, cd);
        PN_STRFREE(tagn);       /* Done with tagname */
        if (error != 0) {
            data->status = tbs_errno2status(error);
            goto out;
        }
    } else {
        mvfs_log(MFS_LOG_DEBUG,
                 "mvfs_setprocview: mfs_copyin_strbuf returned %d"
                 " (str.s=%p str.l=%lu vrtype=%d)\n",
                 error, strbuf->s, strbuf->l, vrtype);
        error = 0; /* Ignore this error (see above). */
        vw = NULL; /* Unset current view */
    }
    /*
     * All looks OK (we only get here with error == 0), now jam the root dir.
     * Note: following call "uses up" a refcount on vw, so we don't VN_RELE it
     * in this level.
     */
    error = MVFS_SET_PROCVIEW(vw, &data->status, cd);

out:

    return error;
}

MVFS_IOCTL_HANDLER(get_procviewinfo)
{
    VNODE_T *cdir;
    VNODE_T *vw;
    mvfs_viewinfo_t viewinfo;
    int  error = 0;

    if ((error = CopyInMvfs_viewinfo(data->infop,
                                     &viewinfo,
                                     callinfo)) != 0)
    {
        return error;
    }

    cdir = MVFS_GET_PROCVIEW(); /* Get process view dir */
    if (cdir == NULL || !MFS_VPISMFS(cdir)) {
        error = ENOENT; /* Not in a view */
    } else {
        vw = MFS_VIEW(cdir); /* Get view */
        if (vw == NULL) { /* No view */
            error = ENOENT;
        } else {
            VN_HOLD(vw); /* Hold view for safety */
            error = mvfs_copyout_viewinfo(&viewinfo, vw);
            ATRIA_VN_RELE(vw, cd);

            if (error == 0) {
                error = CopyOutMvfs_viewinfo(&viewinfo,
                                             data->infop,
                                             callinfo);
            }
        }
    }

    return error;
}

MVFS_IOCTL_HANDLER(get_xattr)
{
    int  error = 0;
    mvfs_io_xattr_t io_xattr;
    VNODE_T *vp = NULL;

    if ((error = mvfs_ioctl_lookup(vrtype,
                                   MCB_OBJPN(data),
                                   mvfs_get_ilflags_from_cb(data),
                                   &vp,
                                   NULL,
                                   cd,
                                   callinfo)) != 0)
    {
        return error;
    }

    if (!MFS_VPISMFS(vp) || !MFS_ISVOB(VTOM(vp))) {
       data->status = TBS_ST_NOT_AN_OBJ;
       goto cleanup;
    }

    if ((error = CopyInMvfs_io_xattr(data->infop,
                                     &io_xattr,
                                     callinfo)) != 0)
    {
        goto cleanup;
    }

    data->status = TBS_ST_OK;

    MLOCK(VTOM(vp));

    switch (io_xattr.xattr_type) {
      case MVFS_IO_XATTR_NULL:
        io_xattr.xvalue = 0;
        break;
      case MVFS_IO_XATTR_MTYPE:
        io_xattr.xvalue = (u_long)VTOM(vp)->mn_vob.attr.mtype;
        break;
      case MVFS_IO_XATTR_XMODE:
        io_xattr.xvalue =
                     (VTOM(vp)->mn_vob.attr.fstat.mode & TBS_FMODE_AUDITED_OBJ);
        break;
      case MVFS_IO_XATTR_NTFILEATTRS:
        io_xattr.xvalue = 0; /* No such attrs yet */
        break;
      default:
        data->status = TBS_ST_EINVAL;
        break;
    }

    MUNLOCK(VTOM(vp));

    if (data->status ==  TBS_ST_OK) {
        error = CopyOutMvfs_io_xattr(&io_xattr, data->infop, callinfo);
    }

cleanup:
    if (vp != NULL) {
        ATRIA_VN_RELE(vp, cd);
    }

    return error;
}

MVFS_IOCTL_HANDLER(set_xattr)
{
    int  error = 0;
    mvfs_io_xattr_t io_xattr;
    CALL_DATA_T *ncd = NULL;
    VNODE_T *vp = NULL;

    if ((error = mvfs_ioctl_lookup(vrtype,
                                   MCB_OBJPN(data),
                                   mvfs_get_ilflags_from_cb(data),
                                   &vp,
                                   NULL,
                                   cd,
                                   callinfo)) != 0)
    {
        return error;
    }

    if (!MFS_VPISMFS(vp) || !MFS_ISVOB(VTOM(vp))) {
        data->status = TBS_ST_NOT_AN_OBJ;
        goto cleanup;
    }

    if ((error = CopyInMvfs_io_xattr(data->infop, &io_xattr, callinfo)) != 0) {
        goto cleanup;
    }

    ncd = MVFS_ALLOC_SUBSTITUTE_CRED(cd,
                                     MVFS_VIEW_CREDS(vp,
                                                     MVFS_CD2CRED(cd),
                                                     FALSE));

    if (!MVFS_SUBSTITUTE_CRED_IS_VALID(ncd)) {
        error = ENOMEM;
        goto cleanup;
    }

    data->status = TBS_ST_OK;

    switch (io_xattr.xattr_type) {
      case MVFS_IO_XATTR_NULL:
        io_xattr.xvalue = 0;
        break;
      case MVFS_IO_XATTR_MTYPE:
        error = mfs_clnt_change_mtype(vp,
                                      (vob_mtype_t)io_xattr.xvalue,
                                      &data->status,
                                      ncd);
        break;
      case MVFS_IO_XATTR_XMODE:
        if (io_xattr.xvalue != TBS_FMODE_AUDITED_OBJ) {
            data->status = TBS_ST_EINVAL;
            break;
        }

        if (!MVFS_ISVTYPE(vp, VREG)) {
            break;
        }

        MLOCK(VTOM(vp));
        error = mvfs_clnt_setattr_locked(vp,
                                         NULL,
                                         io_xattr.xvalue,
                                         MFS_USE_PROCBH,
                                         0,
                                         ncd,
                                         0);
        MUNLOCK(VTOM(vp));
        break;
      case MVFS_IO_XATTR_NTFILEATTRS:
        break;
      default:
        data->status = TBS_ST_EINVAL;
        break;
    }

    if (data->status ==  TBS_ST_OK) {
        error = CopyOutMvfs_io_xattr(&io_xattr,
                                     data->infop,
                                     callinfo);
    }

cleanup:
    if (ncd != NULL) {
        MVFS_FREE_SUBSTITUTE_CRED(ncd);
    }

    if (vp != NULL) {
        ATRIA_VN_RELE(vp, cd);
    }

    return error;
}

MVFS_IOCTL_HANDLER(mount)
{
    int error = 0;
    VNODE_T *vp;

    /* An automatic view VOB mount (e.g., from mount_vob in
    ** rvagent_mvfs_ioctl.cxx) passes in the viewtag under which the VOB should
    ** be mounted as the pn in the data.  Everybody else (e.g.,
    ** mntrpc_mount1_tag in mntrpc_mount.c) passes in NULL for the pn since
    ** their mounts are "global".  The mvfs_ioctl_lookup routine handles a NULL
    ** pn by returning vp = NULL and EINVAL.
    */
    if (vrtype == AUVWROOT) {
        if ((error = mvfs_ioctl_lookup(vrtype,
                                       MCB_OBJPN(data),
                                       MVFS_IL_FLAG_PN_IS_VIEWTAG,
                                       &vp,
                                       NULL,
                                       cd,
                                       callinfo)) != 0)
        {
            goto done;
        }
    } else {
        ASSERT((vrtype == DYVWROOT) && (MCB_OBJPN(data).s == NULL));
        vp = NULL;
    }
    error = mvfs_register_mount(vrtype,
                                vp,
                                data->infop,
                                data->infolen,
                                cd,
                                callinfo);

    if (vp != NULL) {
        ATRIA_VN_RELE(vp, cd);
    }
  done:
    return error;
}

MVFS_IOCTL_HANDLER(unmount)
{
    int error = 0;
    struct mvfs_unmount_info unmount_info;
    VNODE_T *vp;

    BZERO(&unmount_info, sizeof(unmount_info));

    if ((error = CopyInMvfs_unmount_info(data->infop,
                                         &unmount_info,
                                         callinfo)) != 0)
    {
        goto done;
    }
    /* See mount above for the use of pn in the data. */
    if (vrtype == AUVWROOT) {
        if ((error = mvfs_ioctl_lookup(vrtype,
                                       MCB_OBJPN(data),
                                       MVFS_IL_FLAG_PN_IS_VIEWTAG,
                                       &vp,
                                       NULL,
                                       cd,
                                       callinfo)) != 0)
        {
            goto done;
        }
    } else {
        ASSERT((vrtype == DYVWROOT) && (MCB_OBJPN(data).s == NULL));
        vp = NULL;
    }
    error = mvfs_unregister_mount(vrtype, vp, &unmount_info, cd);

    if (vp != NULL) {
        ATRIA_VN_RELE(vp, cd);
    }
  done:
    return error;
}

MVFS_IOCTL_HANDLER(rmallviewtags)
{
    int  error;
    VNODE_T *dvp = NULL;
    struct uio uio;
    IOVEC_T iov;
    caddr_t dirbufp;
    KDIRENT_T *direntp;
    size_t direntlen = KDIRENT_RECLEN(MAXNAMELEN);
    int eofp;

    dirbufp = KMEM_ALLOC(direntlen, KM_SLEEP);
    if (dirbufp == NULL) {
        return ENOMEM;
    }

    uio.uio_iov = &iov;
    MVFS_UIO_OFFSET(&uio) = 0;

    if ((dvp = MVFS_GETVIEWROOT(vrtype)) == NULL) {
        KMEM_FREE(dirbufp, direntlen);
        error = ENXIO;
    } else {
        mfs_viewdiropen(&dvp, 0, MVFS_CD2CRED(cd));
        error = 0;
        while (!error) {
            mfs_uioset(&uio, dirbufp, direntlen,
                       MVFS_UIO_OFFSET(&uio), UIO_SYSSPACE);
            error = mvfs_viewdirreaddir(dvp, &uio, cd, &eofp);
            if ((error) || (uio.uio_resid == direntlen)) {
                break;
            }

            direntp = (KDIRENT_T *)dirbufp;
            while (((caddr_t) direntp) <
                   (dirbufp + (direntlen - uio.uio_resid)))
            {
                if (STRCMP(KDIRENT_GET_NAME(direntp),
                           MVFS_SPECDEV) != 0)
                {
                    mfs_viewdirrmdir(dvp, KDIRENT_GET_NAME(direntp), cd);
                }

                direntp = (KDIRENT_T *)(((caddr_t)(direntp)) +
                                        KDIRENT_GET_RECLEN(direntp));
            }
        }

        mfs_viewdirclose(dvp, 0, 0, MVFS_CD2CRED(cd));
        KMEM_FREE(dirbufp, direntlen);
        ATRIA_VN_RELE(dvp, cd);
    }

    return error;
}

MVFS_IOCTL_HANDLER(unmountall)
{
    int error = 0;
    VNODE_T *vp = NULL;

    if (MCB_OBJPN(data).s == NULL) {
        if (vrtype == AUVWROOT) {
            return EINVAL;
        }
    } else {
        if ((error = mvfs_ioctl_lookup(vrtype,
                                       MCB_OBJPN(data),
                                       (vrtype == AUVWROOT ?
                                        MVFS_IL_FLAG_PN_IS_VIEWTAG :
                                        mvfs_get_ilflags_from_cb(data)),
                                       &vp,
                                       NULL,
                                       cd,
                                       callinfo)) != 0)
        {
            return error;
        }
    }

    if (vrtype == AUVWROOT &&
        ((vp == NULL) ||
         (MFS_VPISMFS(vp) || !MFS_ISVIEW(VTOM(vp)))))
    {
        error = EINVAL;
    } else {
        error = mvfs_unregister_all_vobs(vrtype, vp, cd);
    }

    if (vp != NULL) {
        ATRIA_VN_RELE(vp, cd);
    }

    return error;
}

MVFS_IOCTL_HANDLER(get_poolmaps)
{
    int  error = 0;
    mvfs_ioget_poolmaps_t pmap;
    register int i;
    mfs_strbufpn_t buftmp;
    struct mfs_mntinfo *mmi;
    VFS_T *vfsp;
    /* Split pools are only supported for Dynamic Views. */
    mvfs_vfs_data_t *mvdp = MDKI_VFS_GET_DATAP(NULL);

    if (vrtype == AUVWROOT) {
        return EINVAL;
    }

    if ((error = CopyInMvfs_ioget_poolmaps(data->infop,
                                           &pmap,
                                           callinfo)) != 0)
    {
        return error;
    }

    MVFS_LOCK(&(mvdp->mount_lock));
    vfsp = mvfs_findvfs_uuid(mvdp, &pmap.replica_uuid);
    if (vfsp == NULL) {
        MVFS_UNLOCK(&(mvdp->mount_lock));
        data->status = TBS_ST_NOT_FOUND;
        return 0;
    }

    mmi = VFS_TO_MMI(vfsp);
    if (pmap.mapcount == 0) {
        /* query for how many slots needed */
        pmap.mapcount = mmi->mmi_sptable_len;
        MVFS_UNLOCK(&(mvdp->mount_lock));
        /*
         * Explicitly copy the number of slots back out here
         */
        error = CopyOutMvfs_ioget_poolmaps(&pmap,
                                           data->infop,
                                           callinfo);
        return error;
    }

    /*
     * actually wants the path translations--
     * go do the copyout
     */
    for (i = 0;
         i < KS_MIN(pmap.mapcount, mmi->mmi_sptable_len);
         i++)
    {
        /*
         * only copy out the pattern and
         * the user-version of the
         * pathnames
         */
        error = CopyInMfs_strbufpn_index((caddr_t)pmap.patterns,
                                         &buftmp,
                                         i,
                                         callinfo);
        if (error != 0) {
            break;
        }

        error = mfs_copyout_strbufpn(buftmp,
                                     mmi->mmi_sptable[i].sp_prefix);
        if (error != 0) {
            break;
        }

        error = CopyOutMfs_strbufpn_index(&buftmp,
                                          (caddr_t)pmap.patterns,
                                          i,
                                          callinfo);
        if (error != 0) {
            break;
        }

        error = CopyInMfs_strbufpn_index((caddr_t)pmap.replacements,
                                         &buftmp,
                                         i,
                                         callinfo);
        if (error != 0) {
            break;
        }

        error = mfs_copyout_strbufpn(buftmp,
                                     mmi->mmi_sptable[i].sp_usertarget);
        if (error != 0) {
            break;
        }

        error = CopyOutMfs_strbufpn_index(&buftmp,
                                          (caddr_t)pmap.replacements,
                                          i,
                                          callinfo);
        if (error != 0) {
            break;
        }
    }

    MVFS_UNLOCK(&(mvdp->mount_lock));

    return error;
}

MVFS_IOCTL_HANDLER(unsupported)
{
    mvfs_log(MFS_LOG_INFO,
             "mvfs_ioctl_unsupported: ioctl %d called.\n",
             MCB_CMD(data));

    return ENOSYS;
}

MVFS_IOCTL_HANDLER(abort)
{
    int error = 0;
    extern int mdb_crash;

#if defined(MVFS_DEBUG) || defined(MVFS_CRASH_DEBUG)
    if (MDKI_INGLOBALZONE() && MDKI_SUSER(MVFS_CD2CRED(cd))) {
        mdb_crash = 1;
        error = ENOSYS;         /* We'll crash, but compiler doesn't know. */
    } else {
        MDKI_SET_U_ERROR(0); /* Eliminate suser side effect. */
        error = EPERM;
    }
#else
    mvfs_log(MFS_LOG_INFO,
             "mvfs_ioctl_abort (%s): MVFS not compiled with MVFS_DEBUG, abort "
             "ioctl ignored.\n",
             MVFS_CODE_LOC);

    error = ENOSYS;
#endif

    return error;
}

MVFS_IOCTL_HANDLER(exportviewtag)
{
    int  error = 0;
    mvfs_export_viewinfo_t export_viewinfo;
    VNODE_T *dvp = NULL;
    VNODE_T *vw = NULL;
    char *tagn = NULL;

    if (!MDKI_EXPORT_PRIV(MVFS_CD2CRED(cd))) {
        error = EPERM;
        MDKI_SET_U_ERROR(0);  /* Eliminate suser side effect. */
    } else if ((error = CopyInMvfs_export_viewinfo(data->infop,
                                                   &export_viewinfo,
                                                   callinfo)) == 0)
    {
        /* Get parameters */
        if ((error = mfs_copyin_strbuf(export_viewinfo.viewtag, &tagn)) == 0) {
            if (mvfs_hmname_type(vrtype, tagn, NULL)) {
                data->status = TBS_ST_EINVAL;
            } else if (!MFS_VIDINRANGE(export_viewinfo.exportid)) {
                data->status = TBS_ST_EINVAL;
            } else if ((dvp = MVFS_GETVIEWROOT(vrtype)) == NULL) {
                error = ENXIO;
            } else {
                /* See if existing view-tag */
                if ((error = mvfs_viewtaglookup(vrtype, tagn, &vw, cd)) == 0)
                {
                    /* do view export stuff */
                    error = mvfs_viewdirexport(dvp, vw,
                                               export_viewinfo.exportid);
                    ATRIA_VN_RELE(vw, cd);
                }

                ATRIA_VN_RELE(dvp, cd);
            }

            PN_STRFREE(tagn);
        }
    }

    return error;
}

MVFS_IOCTL_HANDLER(unexportviewtag)
{
    int  error = 0;
    mvfs_viewtag_info_t viewtag_info;
    VNODE_T *dvp = NULL;
    VNODE_T *vp;
    char *tagn = NULL;

    if (!MDKI_EXPORT_PRIV(MVFS_CD2CRED(cd))) {
        error = EPERM;
        MDKI_SET_U_ERROR(0);   /* Eliminate suser side effect */
    } else if ((error = CopyInMvfs_viewtag_info(data->infop,
                                                &viewtag_info,
                                                callinfo)) == 0)
    {
        error = mfs_copyin_strbuf(viewtag_info.viewtag, &tagn);
        if (error == 0 ) {
            /* Get held vnode ptr to viewroot dir */
            if ((dvp = MVFS_GETVIEWROOT(vrtype)) == NULL) {
                error = ENXIO;
            } else {
                if ((error = mvfs_viewtaglookup(vrtype, tagn, &vp, cd)) == 0) {
                    error = mvfs_viewdirunexport(dvp, vp);
                    ATRIA_VN_RELE(vp, cd);
                }

                ATRIA_VN_RELE(dvp, cd);
            }

            PN_STRFREE(tagn);
        }
    }

    return error;
}

MVFS_IOCTL_HANDLER(zero_stats)
{
    int  error = 0;
    int  cpuid;
    mvfs_stats_data_t *sdp;

    if (MDKI_SUSER(MVFS_CD2CRED(cd))) {
        for (cpuid = 0; cpuid < mvfs_max_cpus; cpuid++) {
             if ((sdp = MDKI_STATS_GET_DATAP(cpuid)) != NULL) {
                sdp->zero_me = TRUE;
             }
        }
    } else {
        error = EPERM;
        MDKI_SET_U_ERROR(0);/* Eliminate suser side effect. */
    }

    return error;
}

MVFS_IOCTL_HANDLER(get_cache_usage)
{
    mvfs_cache_usage_t mvfs_cache_usage;

    mvfs_mn_count(&mvfs_cache_usage);
    mvfs_dnc_count(&mvfs_cache_usage);
    mvfs_rpc_count(&mvfs_cache_usage);

    return CopyOutMvfs_cache_usage(&mvfs_cache_usage,
                                   data->infop,
                                   callinfo);
}

MVFS_IOCTL_HANDLER(set_cache_sizes)
{
    int  error = 0;
    mvfs_cache_sizes_t mvfs_cache_size;

    if ((error = CopyInMvfs_cache_sizes(data->infop,
                                        &mvfs_cache_size,
                                        callinfo)) != 0)
    {
        return error;
    }

    if (MDKI_INGLOBALZONE() && MDKI_SUSER(MVFS_CD2CRED(cd))) {
        if (mvfs_cache_size.version >= MVFS_SETCACHE_VERSION) {
            mvfs_cache_size.version = MVFS_SETCACHE_VERSION; /* return it */
            error = mvfs_mn_setcaches(&mvfs_cache_size, cd);
            if (!error) {
                error = mvfs_dnc_setcaches(&mvfs_cache_size, cd);
            }

            if (!error) {
                error = mvfs_rpc_setcaches(&mvfs_cache_size);
            }

            if (!error) {
                error = mvfs_rddir_cache_setcaches(&mvfs_cache_size);
            }

        } else {
            error = EINVAL;
        }
    } else {
        error = EPERM;
        MDKI_SET_U_ERROR(0);   /* Eliminate suser side effect. */
    }

    return error;
}

MVFS_IOCTL_HANDLER(get_view_stats)
{
    int  error = 0;
    mvfs_viewstats_t *mvwsp = NULL;
    char *tagn = NULL;
    mvfs_statbufs_t *sp = NULL;
    VNODE_T *vw = NULL;
    mvfs_common_data_t *mcdp = MDKI_COMMON_GET_DATAP();

    if ((mvwsp = KMEM_ALLOC(sizeof(*mvwsp), KM_SLEEP)) == NULL) {
        return ENOMEM;
    }

    sp = &(mvwsp->stats);

    if ((error = CopyInMvfs_viewstats(data->infop, mvwsp, callinfo)) != 0) {
        goto cleanup;
    }

    /* Copy out the status of per-view statistics collection.  It is disabled
     * by default.
     */
    error = COPYOUT((caddr_t)&(mcdp->mvfs_pview_stat_enabled),
                    (caddr_t)(mvwsp->mvfs_pview_stat_enabled.s),
                    KS_MIN(mvwsp->mvfs_pview_stat_enabled.m,
                    sizeof(int)));

    if (error != 0) {
        mvfs_log(MFS_LOG_DEBUG,
                 "mvfs_ioctl_get_view_stats: Copyout failed for "
                 "mvfs_pview_stat_enabled. error=%d, status=%d\n",
                 error,
                 tbs_errno2status(error));
        goto cleanup;
    }

    if ((error = mfs_copyin_strbuf(mvwsp->viewtag, &tagn)) != 0) {
        goto cleanup;
    }

    /* See if existing viewtag */
    if ((error = mvfs_viewtaglookup(vrtype, tagn, &vw, cd)) != 0) {
        mvfs_log(MFS_LOG_DEBUG,
                 "mvfs_ioctl_get_viewstats: view tag lookup failed for %s, "
                 "error = %d\n",
                 (tagn == NULL ? "NULL" : tagn),
                 error);
        goto cleanup;
    }

    if (sp->clntstat.s && sp->clntstat.m) {
        error = CopyOutMfs_clntstat(&(VTOM(vw)->mn_view.pvstat->clntstat),
                                    (caddr_t)sp->clntstat.s, sp->clntstat.m,
                                    callinfo);
        if (error != 0) {
            goto cleanup;
        }
    }

    if (sp->acstat.s && sp->acstat.m) {
        error = COPYOUT((caddr_t)&(VTOM(vw)->mn_view.pvstat->acstat),
                        (caddr_t)sp->acstat.s,
                        KS_MIN(sp->acstat.m, sizeof(struct mfs_acstat)));

        if (error != 0) {
            goto cleanup;
        }
    }

    if (sp->dncstat.s &&  sp->dncstat.m) {
        error = COPYOUT((caddr_t)&(VTOM(vw)->mn_view.pvstat->dncstat),
                        (caddr_t)sp->dncstat.s,
                        KS_MIN(sp->dncstat.m, sizeof(struct mfs_dncstat)));
    }

cleanup:
    if (tagn != NULL) {
        PN_STRFREE(tagn);
    }

    if (vw != NULL) {
        ATRIA_VN_RELE(vw, cd);
    }

    if (mvwsp != NULL) {
        KMEM_FREE(mvwsp, sizeof(*mvwsp));
    }

    return error;
}

MVFS_IOCTL_HANDLER(zero_view_stats)
{
    int  error = 0;
    mvfs_zero_viewstat_t mvfs_zero_viewstat;
    char *tagn = NULL;
    VNODE_T *vw = NULL;

    if ((error = CopyInMvfs_zero_viewstat(data->infop,
                                          &mvfs_zero_viewstat,
                                          callinfo)) != 0)
    {
        return error;
    }

    if ((error = mfs_copyin_strbuf(mvfs_zero_viewstat.viewtag,
                                   &tagn)) == 0)
    {
        /* See if existing viewtag */
        error = mvfs_viewtaglookup(vrtype, tagn, &vw, cd);
        PN_STRFREE(tagn);

        if (error == 0 ) {
            if (MDKI_SUSER(MVFS_CD2CRED(cd))) {
                MVFS_PVSTAT_ZERO(vw);
            } else {
                error = EPERM;
                MDKI_SET_U_ERROR(0); /* Eliminate suser side effect. */
            }

            ATRIA_VN_RELE(vw, cd);
        }
    }

    return error;
}

MVFS_IOCTL_HANDLER(sidhost_credmapping)
{
    int  error = 0;
    mvfs_sidhost_cred_t sidhost_cred;

    if ((error = CopyInMvfs_sidhost_cred(data->infop,
                                         &sidhost_cred,
                                         callinfo)) == 0)
    {
        error = MVFS_REGISTER_SIDHOST_CREDMAPS(&sidhost_cred);
    }

    return error;
}

MVFS_IOCTL_HANDLER(delete_sidhost_credmapping)
{
    int  error = 0;
    mvfs_sid_t mvfs_sid;

    if ((error = CopyInMvfs_sid(data->infop, &mvfs_sid, callinfo)) == 0) {
        error = MVFS_UNREGISTER_SIDHOST_CREDMAPS(&mvfs_sid);
    }

    return error;
}

MVFS_IOCTL_HANDLER(get_viewtag_export)
{
    int  error;
    VNODE_T *vp;
    mvfs_export_viewinfo_t export_viewinfo;
    char *tagn = NULL;

    if ((error = CopyInMvfs_export_viewinfo(data->infop,
                                            &export_viewinfo,
                                            callinfo)) != 0)
    {
        return error;
    }

    /* Get parameters */
    if ((error = mfs_copyin_strbuf(export_viewinfo.viewtag,
                                   &tagn)) == 0)
    {
        if ((error = mvfs_viewtaglookup(vrtype, tagn, &vp, cd)) == 0) {
            if (VTOM(vp)->mn_view.exid == (u_int) -1) {
                data->status = TBS_ST_NOT_FOUND;
            } else {
                export_viewinfo.exportid = VTOM(vp)->mn_view.exid;
                error = CopyOutMvfs_export_viewinfo(&export_viewinfo,
                                                    data->infop,
                                                    callinfo);
            }

            ATRIA_VN_RELE(vp, cd);
        }

        PN_STRFREE(tagn);
    }

    return error;
}

MVFS_IOCTL_HANDLER(get_gfsinfo)
{
     int error = 0;
     mvfs_gfsinfo_t gfsinfo;

     if ((error = CopyInMvfs_gfsinfo(data->infop, &gfsinfo, callinfo)) == 0) {
         /* Do the actual work of this ioctl if everything is OK. */
         if (error == 0) {
             error = MVFS_GET_GFSINFO(&gfsinfo);
         }

         if (error == 0) {
             error = CopyOutMvfs_gfsinfo(&gfsinfo, data->infop, callinfo);
         }
     }

     return error;
}

MVFS_IOCTL_HANDLER(set_vobrt_vfsmnt)
{
    int  error = 0;
    mfs_strbufpn_pair_t vob_path;

    if ((error = CopyInMfs_strbufpn_pair(data->infop,
                                         &vob_path,
                                         callinfo)) == 0)
    {
        /* We are going to use the pathname strings, so copy them in. */
        MFS_STRBUFPN_PAIR_COPYIN_STRS(&vob_path, error);
        /* Do the actual work of this ioctl if everything is OK. */
        if (error == 0) {
            error = MVFS_SET_VOBRT_VFSMNT(&vob_path);
        }

        /* Cleanup after success or failure (this handles NULL pointers). */
        PNPAIR_STRFREE(&vob_path);
    }

    return error;
}

MVFS_IOCTL_HANDLER(get_cache_sizes)
{
    mvfs_common_data_t *mcdp = MDKI_COMMON_GET_DATAP();
    int  error = 0;
    mvfs_cache_sizes_t mvfs_cache_size;

    if ((error = CopyInMvfs_cache_sizes(data->infop,
                                        &mvfs_cache_size,
                                        callinfo)) != 0)
    {
        return error;
    }

    if (mvfs_cache_size.version >= MVFS_SETCACHE_VERSION) {
        mvfs_cache_size.version = MVFS_SETCACHE_VERSION; /* return it */
        mvfs_cache_size.size[MVFS_SETCACHE_LARGEINIT] = mcdp->mvfs_largeinit;
        mvfs_cache_size.mask = MVFS_SETCACHE_ALLBITS;
        error = mvfs_mn_getcaches(&mvfs_cache_size);
        if (!error) {
            error = mvfs_dnc_getcaches(&mvfs_cache_size);
        }

        if (!error) {
            error = mvfs_rpc_getcaches(&mvfs_cache_size);
        }

        if (!error) {
            error = mvfs_proc_getcaches(&mvfs_cache_size);
        }

        if (!error) {
            error = mvfs_rddir_cache_getcaches(&mvfs_cache_size);
        }

    } else {
        error = EINVAL;
    }

    if (error == 0) {
        error = CopyOutMvfs_cache_sizes(&mvfs_cache_size,
                                        data->infop,
                                        callinfo);
    }

    return error;
}

MVFS_IOCTL_HANDLER(compute_cache_defaults)
{
    int  error = 0;
    ks_int32_t scale_factor;
    mvfs_cache_sizes_t mvfs_cache_size;
    mvfs_common_data_t *mcdp = MDKI_COMMON_GET_DATAP();

    if ((error = CopyInMvfs_cache_sizes(data->infop,
                                        &mvfs_cache_size,
                                        callinfo)) != 0)
    {
        return error;
    }

    if (mvfs_cache_size.version >= MVFS_SETCACHE_VERSION) {
        mvfs_cache_size.version = MVFS_SETCACHE_VERSION; /* return it */
        /* capture requested scale factor */
        scale_factor = mvfs_cache_size.size[MVFS_SETCACHE_LARGEINIT];
        if (scale_factor == MVFS_SETCACHE_DEFAULT_VALUE) {
            mvfs_cache_size.size[MVFS_SETCACHE_LARGEINIT] =
                mcdp->mvfs_computed_largeinit;
            mvfs_cache_size.mask |= MVFS_CACHEBIT(LARGEINIT);
            scale_factor = mcdp->mvfs_computed_largeinit;
        }

        error = mvfs_mn_compute_caches(scale_factor, &mvfs_cache_size);
        if (!error) {
            error = mvfs_dnc_compute_caches(scale_factor, &mvfs_cache_size);
        }

        if (!error) {
            error = mvfs_rpc_compute_caches(scale_factor, &mvfs_cache_size);
        }

        if (!error) {
            error = mvfs_proc_compute_caches(scale_factor, &mvfs_cache_size);
        }

        if (!error) {
            error = mvfs_rddir_compute_caches(scale_factor, &mvfs_cache_size);
        }
    } else {
        error = EINVAL;
    }

    if (error == 0) {
        error = CopyOutMvfs_cache_sizes(&mvfs_cache_size,
                                        data->infop,
                                        callinfo);
    }

    return error;
}

/* This routine called via MVFS_CMD_ENABLE_PVIEW_STATS IOCTL is used to enable
 * per-view statistics collection.  It was found that these statistics are
 * rarely used and more importantly the lock "mvfs_pvac_statlock", protecting
 * mfs_acstat sub-structure in per-view statistics structure is one of the top
 * contended locks.  Hence, per-view statistics collection is disabled by
 * default to improve performance on multi-core, multi-processor systems.
 */
MVFS_IOCTL_HANDLER(enable_pview_stats)
{
    int  error = 0;
    int vnum;
    VNODE_T *vwroot, *vw;
    struct mvfs_pvstat *pvp;
    register mfs_mnode_t *vwrootmnp;
    mvfs_common_data_t *mcdp = MDKI_COMMON_GET_DATAP();

    /* Error out if per-view statistics collection is already enabled. */
    if (mcdp->mvfs_pview_stat_enabled == TRUE) {
        mvfs_log(MFS_LOG_DEBUG, "pview stat already enabled.\n");
        return EINVAL;
    }

    /* Enable per-view statistics */
    mcdp->mvfs_pview_stat_enabled = TRUE;

    /* Need to zero out the per-view statistics collected in all
     * the active views.  It is done with the viewroot mnode locked.
     * This should not be a problem as users would be enabling per-view
     * statistics rarely.
     */
    vwroot = MVFS_GETVIEWROOT(vrtype);
    if (!vwroot) {
        /* If we did not find the view root then cannot zero out
         * the collected per-view statistics.  Hence, disabling per-view
         * stat and returning an error.
         */
        mvfs_log(MFS_LOG_ERR,
                 "Failed to enable per-view statistics: View root not found\n");
        mcdp->mvfs_pview_stat_enabled = FALSE;
        return ENOENT;
    }

    vwrootmnp = VTOM(vwroot);
    MLOCK(vwrootmnp);

    for (vnum = 0; vnum < vwrootmnp->mn_ramdir.hwm; vnum++) {
        if (vwrootmnp->mn_ramdir.ents[vnum].nm == NULL) {
            continue;
        }

        vw = vwrootmnp->mn_ramdir.ents[vnum].vp;

        /* Skip if this is not a view vnode */
        if (MFS_VPISMFS(vw) && MFS_ISVIEW(VTOM(vw))) {
            MFS_HOLDVW(vw);
            pvp = VTOM(vw)->mn_view.pvstat;
            mvfs_pview_stat_zero(pvp);
            ATRIA_VN_RELE(vw, cd);
        }
    }

    MUNLOCK(vwrootmnp);
    ATRIA_VN_RELE(vwroot, cd);
    error = 0;

    return error;
}

/* This routine called via the MVFS_CMD_DISABLE_PVIEW_STATS IOCTL is used to
 * disable the per-view statistics collection if it was enabled using the
 * MVFS_CMD_ENABLE_PVIEW_STATS IOCTL.
 */
MVFS_IOCTL_HANDLER(disable_pview_stats)
{
    int  error = 0;
    mvfs_common_data_t *mcdp = MDKI_COMMON_GET_DATAP();

    /* Disable per-view statistics collection if it is enabled. */
    if (mcdp->mvfs_pview_stat_enabled == TRUE) {
        mcdp->mvfs_pview_stat_enabled = FALSE;
        error = 0;
    } else {
        error = EINVAL;
        mvfs_log(MFS_LOG_DEBUG, "Pview stat already disabled.\n");
    }

    return error;
}

STATIC void
mvfs_addup_clntstat(
    mvfs_stats_data_t *percpu_sdp,
    mvfs_stats_data_t *sdp
)
{

#define ADDUP_FIELD(field) sdp->mfs_clntstat.field +=  \
                                percpu_sdp->mfs_clntstat.field

    ADDUP_FIELD(clntget);
    ADDUP_FIELD(clntfree);
    ADDUP_FIELD(clntcreate);
    ADDUP_FIELD(clntdestroy);
    ADDUP_FIELD(clntcalls);
    ADDUP_FIELD(clntretries);

    ADDUP_FIELD(mfscall);
    ADDUP_FIELD(mfsfail);
    ADDUP_FIELD(mfsintr);
    ADDUP_FIELD(mfsmaxdelay);
    sdp->mfs_clntstat.mfsmaxdelaytime =
        KS_MAX(sdp->mfs_clntstat.mfsmaxdelaytime,
               percpu_sdp->mfs_clntstat.mfsmaxdelaytime);

    ADDUP_FIELD(lpccall);
    ADDUP_FIELD(lpcfail);
    ADDUP_FIELD(lpcmaxdelay);
    sdp->mfs_clntstat.lpcmaxdelaytime =
        KS_MAX(sdp->mfs_clntstat.lpcmaxdelaytime,
               percpu_sdp->mfs_clntstat.lpcmaxdelaytime);

    mvfs_add_times(&(sdp->mfs_clntstat.mvfsthread_time),
                   &(percpu_sdp->mfs_clntstat.mvfsthread_time));
    return;

#undef ADDUP_FIELD
}

STATIC void
mvfs_addup_mnstat(
    mvfs_stats_data_t *percpu_sdp,
    mvfs_stats_data_t *sdp
)
{
#define ADDUP_FIELD(field) sdp->mfs_mnstat.field +=  \
                                percpu_sdp->mfs_mnstat.field

    ADDUP_FIELD(mnget);
    ADDUP_FIELD(mnfound);
    ADDUP_FIELD(mnfoundstale);
    ADDUP_FIELD(mnreusefree);
    ADDUP_FIELD(mncreate);
    ADDUP_FIELD(mngetnum);
    ADDUP_FIELD(mnfree);
    ADDUP_FIELD(mndestroy);
    ADDUP_FIELD(mnreclaim);
    ADDUP_FIELD(mnvobhashcnt);
    ADDUP_FIELD(mncvphashcnt);
    ADDUP_FIELD(mnotherhashcnt);
    ADDUP_FIELD(mnflushvfscnt);
    ADDUP_FIELD(mnflushvwcnt);

    return;

#undef ADDUP_FIELD
}

STATIC void
mvfs_addup_clearstat(
    mvfs_stats_data_t *percpu_sdp,
    mvfs_stats_data_t *sdp
)
{
#define ADDUP_FIELD(field) sdp->mfs_clearstat.field  += \
                                percpu_sdp->mfs_clearstat.field

#define ADDUP_TIME(t_val) (mvfs_add_times(&(sdp->mfs_clearstat.t_val), \
                                          &(percpu_sdp->mfs_clearstat.t_val)))
    ADDUP_FIELD(clearget);
    ADDUP_FIELD(clearcreate);
    ADDUP_FIELD(clearraces);
    ADDUP_FIELD(clearcreatraces);
    ADDUP_FIELD(clearread);
    ADDUP_FIELD(clearwrite);

    ADDUP_TIME(clearget_time);
    ADDUP_TIME(clearcreat_time);
    ADDUP_TIME(clearrd_time);
    ADDUP_TIME(clearwr_time);
    ADDUP_TIME(clearopen_time);
    ADDUP_TIME(unclearrd_time);
    ADDUP_TIME(unclearwr_time);
    ADDUP_TIME(unclearget_time);
    ADDUP_TIME(cto_getattr_time);

    ADDUP_FIELD(clearopen);
    ADDUP_FIELD(unclearopen);
    ADDUP_FIELD(cleargetmiss);
    ADDUP_FIELD(clearreclaim);
    ADDUP_FIELD(clearreclaimmiss);
    ADDUP_FIELD(cleargetlkup);

    return;

#undef ADDUP_FIELD
#undef ADDUP_TIME
}

STATIC void
mvfs_addup_rvcstat(
    mvfs_stats_data_t *percpu_sdp,
    mvfs_stats_data_t *sdp
)
{
#define ADDUP_FIELD(field) sdp->mfs_rvcstat.field  += \
                                percpu_sdp->mfs_rvcstat.field

    ADDUP_FIELD(rvc_hits);
    ADDUP_FIELD(rvc_misses);
    ADDUP_FIELD(rvc_misstimo);
    ADDUP_FIELD(rvc_purge);

    return;

#undef ADDUP_FIELD
}

STATIC void
mvfs_addup_dncstat(
    mvfs_stats_data_t *percpu_sdp,
    mvfs_stats_data_t *sdp
)
{
#define ADDUP_FIELD(field) sdp->mfs_dncstat.field  += \
                                percpu_sdp->mfs_dncstat.field

    ADDUP_FIELD(dnc_hits);
    ADDUP_FIELD(dnc_hitdot);
    ADDUP_FIELD(dnc_hitdir);
    ADDUP_FIELD(dnc_hitreg);
    ADDUP_FIELD(dnc_hitnoent);
    ADDUP_FIELD(dnc_hitbhfromnull);
    ADDUP_FIELD(dnc_misses);
    ADDUP_FIELD(dnc_missinvalid);
    ADDUP_FIELD(dnc_missdncgen);
    ADDUP_FIELD(dnc_missevtime);
    ADDUP_FIELD(dnc_missnoenttimedout);
    ADDUP_FIELD(dnc_missbh);
    ADDUP_FIELD(dnc_missnovp);
    ADDUP_FIELD(dnc_add);
    ADDUP_FIELD(dnc_adddir);
    ADDUP_FIELD(dnc_addreg);
    ADDUP_FIELD(dnc_addnoent);
    ADDUP_FIELD(dnc_addbhinvariant);
    ADDUP_FIELD(dnc_addnoop);
    ADDUP_FIELD(dnc_addbh);
    ADDUP_FIELD(dnc_addlong);
    ADDUP_FIELD(dnc_addunlock);
    ADDUP_FIELD(dnc_change);
    ADDUP_FIELD(dnc_remove);
    ADDUP_FIELD(dnc_invalvp);
    ADDUP_FIELD(dnc_flush);
    ADDUP_FIELD(dnc_invalhits);
    ADDUP_FIELD(dnc_flushvw);
    ADDUP_FIELD(dnc_flushvfs);
    ADDUP_FIELD(dnc_invalvw);
    ADDUP_FIELD(dnc_invalnf);
    ADDUP_FIELD(dnc_missdir);
    ADDUP_FIELD(dnc_missreg);
    ADDUP_FIELD(dnc_missnoent);

    return;

#undef ADDUP_FIELD
}

STATIC void
mvfs_addup_acstat(
    mvfs_stats_data_t *percpu_sdp,
    mvfs_stats_data_t *sdp
)
{
#define ADDUP_FIELD(field) sdp->mfs_acstat.field  += \
                                percpu_sdp->mfs_acstat.field

    ADDUP_FIELD(ac_hits);
    ADDUP_FIELD(ac_misses);
    ADDUP_FIELD(ac_updates);
    ADDUP_FIELD(ac_mod);
    ADDUP_FIELD(ac_expmod);
    ADDUP_FIELD(ac_vobmod);
    ADDUP_FIELD(ac_evmiss);
    ADDUP_FIELD(ac_cto);
    ADDUP_FIELD(ac_timo);
    ADDUP_FIELD(ac_genmiss);
    ADDUP_FIELD(ac_newmiss);
    ADDUP_FIELD(ac_lvuthit);
    ADDUP_FIELD(ac_lvutmiss);
    ADDUP_FIELD(ac_rddirhit);
    ADDUP_FIELD(ac_rddirmiss);

    return;

#undef ADDUP_FIELD
}

STATIC void
mvfs_addup_rlstat(
    mvfs_stats_data_t *percpu_sdp,
    mvfs_stats_data_t *sdp
)
{
#define ADDUP_FIELD(field) sdp->mfs_rlstat.field  += \
                                percpu_sdp->mfs_rlstat.field
    ADDUP_FIELD(rl_hits);
    ADDUP_FIELD(rl_misses);

    return;

#undef ADDUP_FIELD
}

STATIC void
mvfs_addup_austat(
    mvfs_stats_data_t *percpu_sdp,
    mvfs_stats_data_t *sdp
)
{
#define ADDUP_FIELD(field) sdp->mfs_austat.field  += \
                                percpu_sdp->mfs_austat.field

#define ADDUP_TIME(t_val) (mvfs_add_times(&(sdp->mfs_austat.t_val), \
                                          &(percpu_sdp->mfs_austat.t_val)))

    ADDUP_FIELD(au_calls);
    ADDUP_FIELD(au_vgetattr);
    ADDUP_FIELD(au_nvgetattr);
    ADDUP_FIELD(au_dupl);

    ADDUP_TIME(au_time);
    ADDUP_TIME(au_settime);
    ADDUP_TIME(au_ioctltime);

    return;

#undef ADDUP_FIELD
#undef ADDUP_TIME
}

STATIC void
mvfs_addup_eacstat(
    mvfs_stats_data_t *percpu_sdp,
    mvfs_stats_data_t *sdp
)
{
#define ADDUP_FIELD(field) sdp->mvfs_eacstat.field  += \
                               percpu_sdp->mvfs_eacstat.field

    ADDUP_FIELD(eac_hits);
    ADDUP_FIELD(eac_misses);
    ADDUP_FIELD(eac_hashcnt);

    return;

#undef ADDUP_FIELD
}
STATIC void
mvfs_addup_vnopcnt(
    mvfs_stats_data_t *percpu_sdp,
    mvfs_stats_data_t *sdp
)
{
    int i;

    for (i = 0; i < MFS_VNOPCNT; i++) {
         sdp->mfs_vnopcnt[i] += percpu_sdp->mfs_vnopcnt[i];
    }

    return;
}

STATIC void
 mvfs_addup_vfsopcnt(
    mvfs_stats_data_t *percpu_sdp,
    mvfs_stats_data_t *sdp
)
{
    int i;

    for (i = 0; i < MFS_VFSOPCNT; i++) {
         sdp->mfs_vfsopcnt[i] += percpu_sdp->mfs_vfsopcnt[i];
    }

    return;
}

STATIC void
mvfs_addup_viewopcnt(
    mvfs_stats_data_t *percpu_sdp,
    mvfs_stats_data_t *sdp
)
{
    int i;

    for (i = 0; i < VIEW_NUM_PROCS; i++) {
         sdp->mfs_viewopcnt[i] += percpu_sdp->mfs_viewopcnt[i];
         sdp->mvfs_rviewopcnt[i] += percpu_sdp->mvfs_rviewopcnt[i];
    }

    return;
}

STATIC void
mvfs_addup_viewophist(
    mvfs_stats_data_t *percpu_sdp,
    mvfs_stats_data_t *sdp
)
{
    int i,j;

    /* histvalrpc and histvallpc (which are just the histogram bucket time
    ** intervals) are set by mvfs_init_viewophist and never change, so there is
    ** no reason to do anything with them here (they were set in our caller).
    ** We need to add up the other counts...
    */
    for (i = 0; i < MFS_NUM_HISTX; i++) {
         sdp->mvfs_viewophist.histrpc[i] +=
                percpu_sdp->mvfs_viewophist.histrpc[i];
         sdp->mvfs_viewophist.histclrrpc[i] +=
                percpu_sdp->mvfs_viewophist.histclrrpc[i];
         sdp->mvfs_viewophist.histlpc[i] +=
                percpu_sdp->mvfs_viewophist.histlpc[i];
         sdp->mvfs_viewophist.histclrlpc[i] +=
                percpu_sdp->mvfs_viewophist.histclrlpc[i];

         for (j = 0; j < VIEW_NUM_PROCS; j++) {
             sdp->mvfs_viewophist.histperrpc[j][i] +=
                 percpu_sdp->mvfs_viewophist.histperrpc[j][i];
             sdp->mvfs_viewophist.histperlpc[j][i] +=
                 percpu_sdp->mvfs_viewophist.histperlpc[j][i];
         }
    }
    return;
}

STATIC void
mvfs_addup_viewoptime(
    mvfs_stats_data_t *percpu_sdp,
    mvfs_stats_data_t *sdp
)
{
    int i;

    for (i = 0; i < VIEW_NUM_PROCS; i++) {
         mvfs_add_times(&(sdp->mfs_viewoptime[i]),
                        &(percpu_sdp->mfs_viewoptime[i]));
         mvfs_add_times(&(sdp->mvfs_rviewoptime[i]),
                        &(percpu_sdp->mvfs_rviewoptime[i]));
    }

    return;
}

STATIC void
mvfs_add_times(
    timestruc_t *sdp_time,
    timestruc_t *percpu_time
)
{
    timestruc_t tmp;

    tmp.tv_nsec = sdp_time->tv_nsec + percpu_time->tv_nsec;
    tmp.tv_sec = sdp_time->tv_sec + percpu_time->tv_sec;

    if (tmp.tv_nsec  >= 1000000000) {
        tmp.tv_nsec -= 1000000000;
        tmp.tv_sec++;
    }

    sdp_time->tv_nsec = tmp.tv_nsec;
    sdp_time->tv_sec = tmp.tv_sec;

    return;
}

/*
 * Routine to zero out per-view statistics.  We don't zero out the whole
 * per-view stat structure because it has our lock in it.
 */
void
mvfs_pview_stat_zero(struct mvfs_pvstat *pvp)
{
        SPL_T s;

        MVFS_PVCLNT_STATLOCK_LOCK(s, pvp);
        BZERO(&(pvp->clntstat), sizeof(struct mfs_clntstat));
        pvp->clntstat.version = MFS_CLNTSTAT_VERS;
        MVFS_PVCLNT_STATLOCK_UNLOCK(s, pvp);

        MVFS_PVAC_STATLOCK_LOCK(s, pvp);
        BZERO(&(pvp->acstat), sizeof(struct mfs_acstat));
        pvp->acstat.version = MFS_ACSTAT_VERS;
        MVFS_PVAC_STATLOCK_UNLOCK(s, pvp);

        MVFS_PVDNC_STATLOCK_LOCK(s, pvp);
        BZERO(&(pvp->dncstat), sizeof(struct mfs_dncstat));
        pvp->dncstat.version = MFS_DNCSTAT_VERS;
        MVFS_PVDNC_STATLOCK_UNLOCK(s, pvp);
}
static const char vnode_verid_mvfs_mioctl_c[] = "$Id:  281dcaed.826e11e5.83f8.44:37:e6:71:22:3b $";
